from abc import ABC, abstractmethod

from ImagingS import IdObject, IdObjectList, Point
from ImagingS.serialization import PropertySerializable


class Transform(PropertySerializable, IdObject, ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def transform(self, origin: Point) -> Point: pass


class TransformGroup(Transform):
    def __init__(self) -> None:
        super().__init__()
        self.children = IdObjectList()

    @property
    def children(self) -> IdObjectList[Transform]:
        return self._children

    @children.setter
    def children(self, value: IdObjectList[Transform]) -> None:
        assert isinstance(value, IdObjectList)
        self._children = value

    def transform(self, origin: Point) -> Point:
        result = origin
        for tr in self.children:
            result = tr.transform(result)
        return result



from __future__ import annotations

from ImagingS.serialization import PropertySerializable


def _hexNopre(i: int) -> str:
    return format(i, 'X')


class Color(PropertySerializable):
    def __init__(self, r: int = 0, g: int = 0, b: int = 0) -> None:
        super().__init__()
        self.r = r
        self.g = g
        self.b = b

    def __eq__(self, obj: Color) -> bool:
        return self.r == obj.r and self.g == obj.g and self.b == obj.b

    def __repr__(self) -> str:
        return f"Color({self.r}, {self.g}, {self.b})"

    def toHex(self) -> str:
        return f"#{_hexNopre(self.r).zfill(2)}{_hexNopre(self.g).zfill(2)}{_hexNopre(self.b).zfill(2)}"

    @staticmethod
    def fromHex(h: str) -> Color:
        h = h.lstrip("#").zfill(6)[0:6]
        return Color(int(h[0:2], 16), int(h[2:4], 16), int(h[4:6], 16))

    @property
    def r(self) -> int:
        return self._r

    @r.setter
    def r(self, value: int) -> None:
        self._r = int(value)

    @property
    def g(self) -> int:
        return self._g

    @g.setter
    def g(self, value: int) -> None:
        self._g = int(value)

    @property
    def b(self) -> int:
        return self._b

    @b.setter
    def b(self, value: int) -> None:
        self._b = int(value)


class Colors:
    Black = Color(0, 0, 0)
    White = Color(255, 255, 255)
    Red = Color(255, 0, 0)
    Blue = Color(0, 0, 255)
    Green = Color(0, 255, 0)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/runner/work/ImagingS/ImagingS/src/ImagingS/Gui/uic/NewDocumentDialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_NewDocumentDialog(object):
    def setupUi(self, NewDocumentDialog):
        NewDocumentDialog.setObjectName("NewDocumentDialog")
        NewDocumentDialog.resize(286, 142)
        self.verticalLayout = QtWidgets.QVBoxLayout(NewDocumentDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.widget = QtWidgets.QWidget(NewDocumentDialog)
        self.widget.setObjectName("widget")
        self.formLayout = QtWidgets.QFormLayout(self.widget)
        self.formLayout.setObjectName("formLayout")
        self.lblWidth = QtWidgets.QLabel(self.widget)
        self.lblWidth.setObjectName("lblWidth")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.lblWidth)
        self.sbxWidth = QtWidgets.QSpinBox(self.widget)
        self.sbxWidth.setMinimum(100)
        self.sbxWidth.setMaximum(3000)
        self.sbxWidth.setSingleStep(50)
        self.sbxWidth.setProperty("value", 600)
        self.sbxWidth.setObjectName("sbxWidth")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.sbxWidth)
        self.lblHeight = QtWidgets.QLabel(self.widget)
        self.lblHeight.setObjectName("lblHeight")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.lblHeight)
        self.sbxHeight = QtWidgets.QSpinBox(self.widget)
        self.sbxHeight.setMinimum(100)
        self.sbxHeight.setMaximum(3000)
        self.sbxHeight.setSingleStep(50)
        self.sbxHeight.setProperty("value", 600)
        self.sbxHeight.setObjectName("sbxHeight")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.sbxHeight)
        self.verticalLayout.addWidget(self.widget)
        self.buttonBox = QtWidgets.QDialogButtonBox(NewDocumentDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(NewDocumentDialog)
        self.buttonBox.accepted.connect(NewDocumentDialog.accept)
        self.buttonBox.rejected.connect(NewDocumentDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(NewDocumentDialog)

    def retranslateUi(self, NewDocumentDialog):
        _translate = QtCore.QCoreApplication.translate
        NewDocumentDialog.setWindowTitle(_translate("NewDocumentDialog", "New Document"))
        self.lblWidth.setText(_translate("NewDocumentDialog", "Width"))
        self.lblHeight.setText(_translate("NewDocumentDialog", "Height"))



import os
from enum import Enum, unique
from typing import Optional, cast

import qtawesome as qta
from PIL import Image
from PyQt5.QtGui import QCloseEvent, QKeySequence
from PyQt5.QtWidgets import (QAbstractItemView, QColorDialog, QDialog,
                             QFileDialog, QMainWindow, QMessageBox, QUndoStack)

import ImagingS.Gui.ui as ui
from ImagingS import Color
from ImagingS.brush import Brush, Brushes, SolidBrush
from ImagingS.document import Document, DocumentFormat, VersionController
from ImagingS.drawing import (Drawing, DrawingGroup, GeometryDrawing,
                              NumpyArrayRenderContext)
from ImagingS.Gui import icons
from ImagingS.Gui.models import (BrushModel, CommitCommand, DrawingModel,
                                 PropertyModel, TransformModel)
from ImagingS.transform import Transform, TransformGroup

from . import (DocumentEditor, DocumentEditorState, NewDocumentDialog,
               VisualPageState)


@unique
class MainWindowState(Enum):
    Disable = 0,
    Visual = 1,
    Code = 2


class MainWindow(QMainWindow, ui.MainWindow):
    def __init__(self):
        super().__init__()
        self.setupUi(self)
        self.setupUndo()
        self.setupDockWidget()
        self.setupStatusBar()
        self.setupEditor()
        self.setupIcon()
        self._file = None
        self._document = None
        self.documentChanged = False

        self.actClose.triggered.connect(self.actClose_triggered)
        self.actAbout.triggered.connect(self.actAbout_triggered)
        self.actQuit.triggered.connect(self.actQuit_triggered)
        self.actNew.triggered.connect(self.actNew_triggered)
        self.actSave.triggered.connect(self.actSave_triggered)
        self.actSaveAs.triggered.connect(self.actSaveAs_triggered)
        self.actOpen.triggered.connect(self.actOpen_triggered)
        self.actExport.triggered.connect(self.actExport_triggered)

        self.actBrushSolid.triggered.connect(self.actBrushSolid_triggered)

        self.actViewVisual.triggered.connect(self.actViewVisual_triggered)
        self.actViewCode.triggered.connect(self.actViewCode_triggered)

        self.modelBrush = BrushModel(self)
        self.trvBrushes.setModel(self.modelBrush)
        self.trvBrushes.clicked.connect(
            self.trvBrushes_clicked)
        self.actBrushRemove.triggered.connect(self.actBrushRemove_triggered)
        self.actBrushClear.triggered.connect(self.actBrushClear_triggered)

        self.modelDrawing = DrawingModel(self)
        self.modelDrawing.changed.connect(self.modelDrawing_changed)
        self.trvDrawings.setModel(self.modelDrawing)
        self.trvDrawings.clicked.connect(
            self.trvDrawings_clicked)
        self.actDrawingRemove.triggered.connect(
            self.actDrawingRemove_triggered)
        self.actDrawingClear.triggered.connect(self.actDrawingClear_triggered)

        self.modelTransform = TransformModel(self)
        self.trvTransforms.setModel(self.modelTransform)
        self.trvTransforms.clicked.connect(
            self.trvTransforms_clicked)
        self.actTransformRemove.triggered.connect(
            self.actTransformRemove_triggered)
        self.actTransformClear.triggered.connect(
            self.actTransformClear_triggered)

        self.actBrushesRefresh.triggered.connect(
            self.actBrushesRefresh_triggered)
        self.actDrawingsRefresh.triggered.connect(
            self.actDrawingsRefresh_triggered)
        self.actTransformsRefresh.triggered.connect(
            self.actTransformsRefresh_triggered)
        self.actPropertiesRefresh.triggered.connect(
            self.actPropertiesRefresh_triggered)

        self.modelProperties = PropertyModel(self)
        self.trvProperties.setModel(self.modelProperties)

        self._freshAll()

    def setupDockWidget(self):
        self.actToggleBrushes = self.dwgBrushes.toggleViewAction()
        self.actToggleBrushes.setShortcut("Ctrl+Shift+B")
        self.actToggleProperties = self.dwgProperties.toggleViewAction()
        self.actToggleProperties.setShortcut("Ctrl+Shift+P")
        self.actToggleDrawings = self.dwgDrawings.toggleViewAction()
        self.actToggleDrawings.setShortcut("Ctrl+Shift+D")
        self.actToggleTransforms = self.dwgTransforms.toggleViewAction()
        self.actToggleTransforms.setShortcut("Ctrl+Shift+T")

        self.tabifyDockWidget(self.dwgBrushes, self.dwgDrawings)
        self.tabifyDockWidget(self.dwgBrushes, self.dwgTransforms)
        self.dwgDrawings.raise_()

        viewActions = [self.actToggleDrawings, self.actToggleBrushes,
                       self.actToggleTransforms, self.actToggleProperties]
        self.mnuView.addActions(viewActions)

    def setupUndo(self) -> None:
        self.versionController = VersionController()

        self.undoStack = QUndoStack(self)
        self.actUndo = self.undoStack.createUndoAction(self, "Undo")
        self.actUndo.setShortcut(QKeySequence.Undo)
        self.actRedo = self.undoStack.createRedoAction(self, "Redo")
        self.actRedo.setShortcut(QKeySequence.Redo)

        self.mnuEdit.addAction(self.actUndo)
        self.mnuEdit.addAction(self.actRedo)

    def setupStatusBar(self) -> None:
        pass

    def setupEditor(self) -> None:
        self.editor = DocumentEditor()
        self.editor.setObjectName("editor")
        self.gridLayout.addWidget(self.editor, 0, 0, 1, 1)
        self.editor.messaged.connect(self.editor_messaged)
        self.editor.documentCommitted.connect(self.editor_documentCommitted)
        self.editor.stateChanged.connect(self.editor_stateChanged)
        self.editor.visual.stateChanged.connect(self.editor_stateChanged)
        self.editor.code.stateChanged.connect(self.editor_stateChanged)

        self.mnuDrawing.addActions([
            self.editor.visual.actDrawingLine,
            self.editor.visual.actDrawingPolyline,
            self.editor.visual.actDrawingPolygon,
            self.editor.visual.actDrawingRectangle,
            self.editor.visual.actDrawingCurve,
            self.editor.visual.actDrawingEllipse,
        ])
        self.mnuDrawing.addSeparator()
        self.mnuDrawing.addActions([
            self.actDrawingRemove,
            self.actDrawingClear
        ])

        self.mnuTransform.addActions([
            self.editor.visual.actTransformTranslate,
            self.editor.visual.actTransformScale,
            self.editor.visual.actTransformRotate,
            self.editor.visual.actTransformSkew,
            self.editor.visual.actTransformMatrix,
            self.editor.visual.actTransformGroup,
        ])
        self.mnuTool.addActions([
            self.editor.visual.actClip
        ])
        self.mnuTransform.addSeparator()
        self.mnuTransform.addActions([
            self.actTransformRemove,
            self.actTransformClear
        ])

    def setupIcon(self):
        self.actNew.setIcon(qta.icon("mdi.file"))
        self.actOpen.setIcon(qta.icon("mdi.folder-open"))
        self.actSave.setIcon(qta.icon("mdi.content-save"))
        self.actSaveAs.setIcon(qta.icon("mdi.content-save-move"))
        self.actExport.setIcon(qta.icon("mdi.export"))
        self.actClose.setIcon(qta.icon("mdi.close"))
        self.actAbout.setIcon(qta.icon("mdi.information"))
        self.actQuit.setIcon(qta.icon("mdi.exit-to-app"))
        self.actUndo.setIcon(qta.icon("mdi.undo"))
        self.actRedo.setIcon(qta.icon("mdi.redo"))
        self.actBrushSolid.setIcon(icons.solidBrush)
        self.actBrushRemove.setIcon(qta.icon("mdi.close", color="red"))
        self.actDrawingRemove.setIcon(qta.icon("mdi.close", color="red"))
        self.actTransformRemove.setIcon(qta.icon("mdi.close", color="red"))
        self.actBrushClear.setIcon(qta.icon("mdi.delete", color="red"))
        self.actDrawingClear.setIcon(qta.icon("mdi.delete", color="red"))
        self.actTransformClear.setIcon(qta.icon("mdi.delete", color="red"))
        self.dwgBrushes.setWindowIcon(icons.brush)
        self.dwgDrawings.setWindowIcon(icons.drawing)
        self.dwgProperties.setWindowIcon(icons.property)
        self.dwgTransforms.setWindowIcon(icons.transform)
        self.actDrawingsRefresh.setIcon(icons.refresh)
        self.actBrushesRefresh.setIcon(icons.refresh)
        self.actPropertiesRefresh.setIcon(icons.refresh)
        self.actTransformsRefresh.setIcon(icons.refresh)
        self.actToggleDrawings.setIcon(icons.drawing)
        self.actToggleBrushes.setIcon(icons.brush)
        self.actToggleProperties.setIcon(icons.property)
        self.actToggleTransforms.setIcon(icons.transform)
        self.actViewCode.setIcon(icons.code)
        self.actViewVisual.setIcon(icons.visual)
        self.setWindowIcon(qta.icon("mdi.pencil-box-multiple", color="purple"))

    def closeEvent(self, event: QCloseEvent) -> None:
        if self.document is not None:
            self.actClose.trigger()
        if self.document is None:
            event.accept()
        else:
            event.ignore()

    @property
    def file(self) -> Optional[str]:
        return self._file

    @file.setter
    def file(self, value: Optional[str]) -> None:
        self._file = os.path.realpath(value) if value else None
        self._freshTitle()

    @property
    def document(self) -> Optional[Document]:
        return self._document

    @document.setter
    def document(self, value: Optional[Document]) -> None:
        self._document = value
        self._freshTitle()

    @property
    def documentChanged(self) -> bool:
        return self._documentChanged

    @documentChanged.setter
    def documentChanged(self, value: bool) -> None:
        self._documentChanged = value
        self._freshTitle()

    def _freshTitle(self) -> None:
        if self.file is None:
            if self.document is None:
                self.setWindowTitle("ImagingS")
            else:
                if self.documentChanged:
                    self.setWindowTitle("Untitled * - ImagingS")
                else:
                    self.setWindowTitle("Untitled - ImagingS")
        else:
            if self.documentChanged:
                self.setWindowTitle(f"{self.file} * - ImagingS")
            else:
                self.setWindowTitle(f"{self.file} - ImagingS")

    def _freshBrushes(self):
        hasDoc = self.document is not None
        self.dwgBrushes.setEnabled(hasDoc)
        self.modelBrush.clear_items()
        if hasDoc:
            for br in self.document.brushes:
                self.modelBrush.append(br)

    def _freshDrawings(self):
        hasDoc = self.document is not None
        self.dwgDrawings.setEnabled(hasDoc)
        if hasDoc:
            self.modelDrawing.fresh(self.document.drawings)
            self.trvDrawings.expandAll()
        else:
            self.modelDrawing.fresh()

    def _freshTransforms(self):
        drawing = self._currentDrawing()
        hasDrawing = drawing is not None
        self.dwgTransforms.setEnabled(hasDrawing)
        trans = None
        if hasDrawing:
            if isinstance(drawing, GeometryDrawing):
                trans = drawing.geometry.transform
            elif isinstance(drawing, DrawingGroup):
                trans = drawing.transform
        self.modelTransform.fresh(trans)
        if trans is not None:
            self.trvTransforms.expandAll()

    def _freshProperties(self):
        hasDoc = self.document is not None
        self.modelProperties.fresh()
        self.dwgProperties.setEnabled(hasDoc)

    def _freshDockWidgets(self):
        self._freshProperties()
        self._freshBrushes()
        self._freshDrawings()
        self._freshTransforms()

    def _freshActions(self):
        hasDoc = self.document is not None

        self.actClose.setEnabled(hasDoc)
        self.actSave.setEnabled(hasDoc)
        self.actSaveAs.setEnabled(hasDoc)
        self.actExport.setEnabled(hasDoc)
        self.actNew.setEnabled(not hasDoc)

        self.mnuEdit.setEnabled(hasDoc)
        self.mnuView.setEnabled(hasDoc)
        self.mnuDrawing.setEnabled(hasDoc)
        self.mnuTransform.setEnabled(hasDoc)
        self.mnuBrush.setEnabled(hasDoc)
        self.mnuTool.setEnabled(hasDoc)

        isVisualNormal = self.editor.state is DocumentEditorState.Visual and self.editor.visual.state is VisualPageState.Normal
        self.actDrawingClear.setEnabled(isVisualNormal)
        self.actBrushClear.setEnabled(isVisualNormal)
        self.actTransformClear.setEnabled(isVisualNormal)
        self.actDrawingRemove.setEnabled(isVisualNormal)
        self.actBrushRemove.setEnabled(isVisualNormal)
        self.actTransformRemove.setEnabled(isVisualNormal)
        self.mnuDrawing.setEnabled(isVisualNormal)
        self.mnuTransform.setEnabled(isVisualNormal)
        self.mnuBrush.setEnabled(isVisualNormal)
        self.mnuTool.setEnabled(isVisualNormal)

        if isVisualNormal:
            tri = cast(QAbstractItemView.EditTriggers,
                       QAbstractItemView.DoubleClicked | QAbstractItemView.EditKeyPressed)
            self.trvTransforms.setEditTriggers(tri)
            self.trvBrushes.setEditTriggers(tri)
            self.trvDrawings.setEditTriggers(tri)
            self.trvProperties.setEditTriggers(tri)
        else:
            self.trvTransforms.setEditTriggers(
                QAbstractItemView.NoEditTriggers)
            self.trvBrushes.setEditTriggers(QAbstractItemView.NoEditTriggers)
            self.trvDrawings.setEditTriggers(QAbstractItemView.NoEditTriggers)
            self.trvProperties.setEditTriggers(
                QAbstractItemView.NoEditTriggers)

    def _freshEditor(self):
        if self.editor.state is not DocumentEditorState.Disable:
            self.editor.disable()

        if self.document is not None:
            self.editor.enable(self.document)

    def _freshAll(self):
        self._freshTitle()
        self._freshDockWidgets()
        self._freshEditor()
        self._freshActions()

    def _commitDocument(self, message: str):
        assert self.document is not None
        self.documentChanged = True
        cmt = self.versionController.commit(self.document, message)
        self.undoStack.push(CommitCommand(
            cmt, self.versionController, self.undoCallback))

    def _currentDrawing(self) -> Optional[Drawing]:
        indexs = self.trvDrawings.selectedIndexes()
        if len(indexs) == 0:
            return None
        return self.modelDrawing.getUserData(indexs[0])

    def _currentBrush(self) -> Optional[Brush]:
        indexs = self.trvBrushes.selectedIndexes()
        if len(indexs) == 0:
            return None
        return self.modelBrush.getUserData(indexs[0])

    def _currentTransform(self) -> Optional[Transform]:
        indexs = self.trvTransforms.selectedIndexes()
        if len(indexs) == 0:
            return None
        return self.modelTransform.getUserData(indexs[0])

    def trvBrushes_clicked(self, index):
        item = self.modelBrush.getUserData(index)
        if self.modelProperties.obj is not item:
            self.editor.visual.brush = item
            self.modelProperties.fresh(item)
            self.trvProperties.expandAll()
        else:
            self.trvBrushes.clearSelection()
            self.editor.visual.brush = None
            self.modelProperties.fresh()

    def trvDrawings_clicked(self, index):
        item = self.modelDrawing.getUserData(index)
        if item is not self.document.drawings and self.modelProperties.obj is not item:
            self.editor.visual.drawing = item
            self._freshTransforms()
            self.modelProperties.fresh(item)
            self.trvProperties.expandAll()
        else:
            self.trvDrawings.clearSelection()
            self.editor.visual.drawing = None
            self._freshTransforms()
            self.modelProperties.fresh()

    def trvTransforms_clicked(self, index):
        item = self.modelTransform.getUserData(index)
        if self.modelProperties.obj is not item:
            self.modelProperties.fresh(item)
            self.trvProperties.expandAll()
        else:
            self.trvTransforms.clearSelection()
            self.modelProperties.fresh()

    def actViewVisual_triggered(self):
        self.editor.switchVisual()

    def actViewCode_triggered(self):
        self.editor.switchCode()

    def actDrawingRemove_triggered(self):
        item = self._currentDrawing()
        if item is None:
            return
        parent = item.parent()
        if parent is None:
            return
        del parent[item.id]
        self._commitDocument(f"Deleted drawing ({item}).")
        self._freshDrawings()
        self.editor.fresh()
        self.stbMain.showMessage(f"Deleted drawing ({item}).")

    def actDrawingClear_triggered(self):
        self.document.drawings.children.clear()
        self._commitDocument("Cleared drawings")
        self._freshDrawings()
        self.editor.fresh()
        self.stbMain.showMessage("Cleared drawings.")

    def actBrushSolid_triggered(self):
        assert self.document is not None
        color = QColorDialog.getColor()
        if not color.isValid():
            return
        br = SolidBrush(Color(
            color.red(), color.green(), color.blue()))
        self.document.brushes.append(br)
        self._commitDocument("Create SolidBrush")
        self._freshBrushes()
        self.stbMain.showMessage(
            f"Add new brush ({br}).")

    def actBrushRemove_triggered(self):
        item = self._currentBrush()
        if item is None:
            return
        parent = item.parent()
        if parent is None:
            return
        del parent[item.id]
        self._commitDocument(f"Deleted brush ({item})")
        self._freshBrushes()
        self.stbMain.showMessage(
            f"Deleted brush ({item}).")

    def actBrushClear_triggered(self):
        self.document.brushes.clear()
        self._commitDocument("Cleared brushes")
        self._freshBrushes()
        self.stbMain.showMessage("Cleared brushes.")

    def actTransformRemove_triggered(self):
        item = self._currentTransform()
        if item is None:
            return
        drawing = self._currentDrawing()
        if isinstance(drawing, GeometryDrawing):
            if item == drawing.geometry.transform:
                drawing.geometry.transform = None
            else:
                parent = item.parent()
                if parent is None:
                    return
                del parent[item.id]
        elif isinstance(drawing, DrawingGroup):
            if item == drawing.transform:
                drawing.transform = None
            elif isinstance(drawing.transform, TransformGroup):
                parent = item.parent()
                if parent is None:
                    return
                del parent[item.id]
        self._commitDocument(f"Deleted transform ({item})")
        self._freshTransforms()
        self.editor.fresh()
        self.stbMain.showMessage(
            f"Deleted transform ({item}).")

    def actTransformClear_triggered(self):
        drawing = self._currentDrawing()
        if drawing is None:
            return
        if isinstance(drawing, GeometryDrawing):
            if isinstance(drawing.geometry.transform, TransformGroup):
                drawing.geometry.transform.children.clear()
            else:
                drawing.geometry.transform = None
        elif isinstance(drawing, DrawingGroup):
            if isinstance(drawing.transform, TransformGroup):
                drawing.transform.children.clear()
            else:
                drawing.transform = None
        self._commitDocument("Cleared transforms")
        self._freshTransforms()
        self.editor.fresh()
        self.stbMain.showMessage("Cleared transforms.")

    def actClose_triggered(self):
        if self.documentChanged:
            code = QMessageBox.warning(self, "Unsaved changes", "Do you want save changes?",
                                       cast(QMessageBox.StandardButtons, QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel), QMessageBox.Yes)
            if code == QMessageBox.Cancel:
                return
            elif code == QMessageBox.Yes:
                self.actSave.trigger()
        self.undoStack.clear()
        self.versionController.clear()
        self.document = None
        self.documentChanged = False
        self.file = None
        self._freshAll()

    def actNew_triggered(self):
        newDialog = NewDocumentDialog()
        if newDialog.exec_() == QDialog.Accepted:
            doc = Document()
            doc.brushes.append(Brushes.Black)
            doc.size = newDialog.documentSize
            self.document = doc
            self.documentChanged = True
            self.versionController.commit(self.document, "Initial")
            self._freshAll()
            self.stbMain.showMessage(
                f"Created new document with size ({doc.size.width}, {doc.size.height}).")

    def actSave_triggered(self):
        file = self.file
        if file is None:
            self.actSaveAs_triggered()
            return
        if file.endswith(".isd.json"):
            with open(file, mode="w+") as f:
                self.document.save(f, DocumentFormat.RAW)
        else:
            with open(file, mode="wb") as f:
                self.document.save(f)
        self.documentChanged = False
        self.stbMain.showMessage(f"Saved to {self.file}.")

    def actSaveAs_triggered(self):
        options = QFileDialog.Options()
        fileName, okPressed = QFileDialog.getSaveFileName(
            self, "Save As", "", "ImagingS Document (*.isd);; ImagingS Raw Document (*.isd.json)", options=options)
        if not okPressed or not fileName:
            return
        fileName = os.path.realpath(fileName)
        self.file = fileName
        self.actSave.trigger()

    def actOpen_triggered(self):
        options = QFileDialog.Options()
        fileName, okPressed = QFileDialog.getOpenFileName(
            self, "Open", "", "ImagingS Document (*.isd);; ImagingS Raw Document (*.isd.json)", options=options)
        if not okPressed or not fileName:
            return
        fileName = os.path.realpath(fileName)
        if fileName.endswith(".isd.json"):
            with open(fileName, mode="r") as f:
                doc = Document.load(f, DocumentFormat.RAW)
        else:
            with open(fileName, mode="rb") as f:
                doc = Document.load(f)
        self.file = fileName
        self.document = doc
        self.documentChanged = False
        self.versionController.commit(self.document, "Initial")
        self._freshAll()
        self.stbMain.showMessage(f"Opend document at {self.file}.")

    def actExport_triggered(self):
        options = QFileDialog.Options()
        fileName, _ = QFileDialog.getSaveFileName(
            self, "Export To", "",
            "PNG Image (*.png);; JPEG Image (*.jpeg);; BMP Image (*.bmp)",
            options=options)
        if fileName:
            fileName = os.path.realpath(fileName)
            ext = os.path.splitext(fileName)[1].lstrip(".")
            doc = self.document
            assert doc is not None
            context = NumpyArrayRenderContext(
                NumpyArrayRenderContext.create_array(doc.size))

            doc.drawings.render(context)

            Image.fromarray(context.array).save(fileName, ext, quality=95)

            self.stbMain.showMessage(f"Exported to {fileName}.")

    def actAbout_triggered(self) -> None:
        QMessageBox.information(self, "About ImagingS", """
Author    StardustDL
Project   https://github.com/StardustDL/ImagingS
License   MPL-2.0
""".strip())

    def actQuit_triggered(self) -> None:
        self.close()

    def actBrushesRefresh_triggered(self) -> None:
        self._freshBrushes()

    def actDrawingsRefresh_triggered(self) -> None:
        self._freshDrawings()

    def actTransformsRefresh_triggered(self) -> None:
        self._freshTransforms()

    def actPropertiesRefresh_triggered(self) -> None:
        self.modelProperties.fresh(self.modelProperties.obj)

    def editor_messaged(self, message: str) -> None:
        self.stbMain.showMessage(message)

    def editor_documentCommitted(self, document: Document, message: str) -> None:
        self.document = document
        self._commitDocument(message)
        self._freshDockWidgets()

    def undoCallback(self, document: Document) -> None:
        self.document = document
        self._freshAll()

    def editor_stateChanged(self) -> None:
        if self.editor.state is DocumentEditorState.Visual:
            self.actViewCode.setChecked(False)
            self.actViewVisual.setChecked(True)
        elif self.editor.state is DocumentEditorState.Code:
            self.actViewCode.setChecked(True)
            self.actViewVisual.setChecked(False)

        self._freshActions()

    def modelDrawing_changed(self) -> None:
        self._freshDrawings()



from __future__ import annotations

from typing import Tuple

from ImagingS.serialization import PropertySerializable


class Size(PropertySerializable):
    def __init__(self, width: float = 0.0, height: float = 0.0) -> None:
        super().__init__()
        self.width = width
        self.height = height

    def __eq__(self, obj: Size) -> bool:
        return self.width == obj.width and self.height == obj.height

    def __repr__(self) -> str:
        return f"Size({self.width}, {self.height})"

    def asTuple(self) -> Tuple[float, float]:
        return self.width, self.height

    @property
    def width(self) -> float:
        return self._width

    @width.setter
    def width(self, value: float) -> None:
        self._width = float(value)

    @property
    def height(self) -> float:
        return self._height

    @height.setter
    def height(self, value: float) -> None:
        self._height = float(value)



from __future__ import annotations

from typing import List

from PyQt5.QtCore import QObject
from PyQt5.QtWidgets import QApplication


class Application(QObject):

    def __init__(self, argv: List[str]):
        super().__init__()
        self._app = QApplication(argv)
        Application._current = self

    @property
    def app(self) -> QApplication:
        return self._app

    def run(self) -> int:
        return self.app.exec_()

    @classmethod
    def current(cls) -> Application:
        return cls._current



from PyQt5.QtCore import QPointF

from ImagingS import Size
from ImagingS.drawing import GeometryDrawing
from ImagingS.geometry import PolylineGeometry
from ImagingS.Gui import converters

from . import GeometryInteractivity


class PolylineInteractivity(GeometryInteractivity[PolylineGeometry]):
    def __init__(self, target: GeometryDrawing, geometry: PolylineGeometry, size: Size) -> None:
        super().__init__(target, geometry, size)

    def start(self) -> None:
        self._hasStarted = False
        super().start()

    def onMouseRelease(self, point: QPointF) -> None:
        if not self._hasStarted:
            self.geometry.vertexes.append(converters.point(point))
            self.geometry.vertexes.append(
                converters.point(point))  # for next vertex
            self._hasStarted = True
        else:
            self.geometry.vertexes.append(
                converters.point(point))  # for next vertex
        super().onMouseRelease(point)

    def onMouseMove(self, point: QPointF) -> None:
        if self._hasStarted:
            self.geometry.vertexes[-1] = converters.point(point)
            self.update()
        super().onMouseMove(point)

    def onMouseDoubleClick(self, point: QPointF) -> None:
        if len(self.geometry.vertexes) > 1:
            self.geometry.vertexes.remove(self.geometry.vertexes[-1])
        self.update()
        self.end(True)
        super().onMouseDoubleClick(point)



from ._Core import feq, fsign
from ._Color import Color, Colors
from ._Point import Point
from ._Size import Size
from ._Rect import Rect, RectMeasurer
from ._IdObject import IdObject, IdObjectList
from ._Cache import Cache

__all__ = (
    "Color",
    "Colors",
    "Point",
    "Rect",
    "Size",
    "IdObjectList",
    "IdObject",
    "Cache",
    "RectMeasurer",
    "feq",
    "fsign",
)



from __future__ import annotations

from typing import Tuple

import numpy as np

from ImagingS.serialization import PropertySerializable
from ImagingS import feq


class Point(PropertySerializable):
    def __init__(self, x: float = 0.0, y: float = 0.0) -> None:
        super().__init__()
        self.x = x
        self.y = y

    def __eq__(self, other: Point) -> bool:
        return feq(self.x, other.x) and feq(self.y, other.y)

    def __add__(self, other: Point) -> Point:
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other: Point) -> Point:
        return Point(self.x - other.x, self.y - other.y)

    def __neg__(self) -> Point:
        return Point(-self.x, -self.y)

    def __pos__(self) -> Point:
        return self.clone()

    def __mul__(self, other: float) -> Point:
        return Point(self.x * other, self.y * other)

    def __rmul__(self, other: float) -> Point:
        return self * other

    def __abs__(self) -> float:
        return (self.x**2 + self.y**2)**0.5

    def __repr__(self) -> str:
        return f"Point({self.x}, {self.y})"

    def clone(self) -> Point:
        return Point(self.x, self.y)

    @property
    def x(self) -> float:
        return self._x

    @x.setter
    def x(self, value: float) -> None:
        self._x = float(value)

    @property
    def y(self) -> float:
        return self._y

    @y.setter
    def y(self, value: float) -> None:
        self._y = float(value)

    def toHomogeneous(self) -> np.ndarray:
        return np.array([[self.x], [self.y], [1]])

    def asTuple(self) -> Tuple[float, float]:
        return self.x, self.y

    @staticmethod
    def fromHomogeneous(arr: np.ndarray) -> Point:
        assert arr.shape == (3, 1)
        w = float(arr[2][0])
        assert w != 0
        return Point(float(arr[0][0]) / w, float(arr[1][0]) / w)



from typing import Any, Iterator


class Property:
    def __init__(self, name: str, onwer, prop: property):
        self._name = name
        self._onwer = onwer
        self._get = prop.fget
        self._set = prop.fset
        self._del = prop.fdel
        # self._dataType = prop.fget.__annotations__["return"]

    @property
    def name(self) -> str:
        return self._name

    @property
    def canGet(self) -> bool:
        return self._get is not None

    @property
    def canSet(self) -> bool:
        return self._set is not None

    @property
    def canDel(self) -> bool:
        return self._del is not None

    def get(self) -> Any:
        return self._get.__call__(self._onwer)

    def set(self, val) -> None:
        return self._set.__call__(self._onwer, val)

    def delete(self) -> None:
        return self._del.__call__(self._onwer)


def getProperties(obj: Any) -> Iterator[Property]:
    cls = obj.__class__
    for name in dir(cls):
        item = getattr(cls, name)
        if isinstance(item, property):
            yield Property(name, obj, item)



from PyQt5.QtCore import QModelIndex, Qt
from PyQt5.QtGui import QStandardItem, QStandardItemModel

from ImagingS.Gui import icons
from ImagingS.transform import (MatrixTransform, RotateTransform,
                                ScaleTransform, SkewTransform, Transform,
                                TransformGroup, TranslateTransform)


class TransformModel(QStandardItemModel):
    TYPE = 0

    def __init__(self, parent) -> None:
        super().__init__(0, 1, parent)
        self.setHeaderData(TransformModel.TYPE, Qt.Horizontal, "Type")
        self.obj = None

    def fresh(self, obj: Transform = None) -> None:
        self.removeRows(0, self.rowCount())
        self.obj = obj
        if obj is None:
            return

        name = obj.__class__.__name__
        item = QStandardItem(name)
        self._setIcon(item, obj)
        self._setUserData(item, obj)
        self.appendRow(item)

        if isinstance(obj, TransformGroup):
            self._addChildren(item, obj)

    def getUserData(self, index: QModelIndex) -> Transform:
        item = self.itemFromIndex(index)
        return item.data(Qt.UserRole)

    def _setUserData(self, item: QStandardItem, value: Transform) -> None:
        item.setData(value, Qt.UserRole)

    def _setIcon(self, item: QStandardItem, value: Transform) -> None:
        if isinstance(value, TranslateTransform):
            item.setIcon(icons.translateTransform)
        elif isinstance(value, SkewTransform):
            item.setIcon(icons.skewTransform)
        elif isinstance(value, RotateTransform):
            item.setIcon(icons.rotateTransform)
        elif isinstance(value, ScaleTransform):
            item.setIcon(icons.scaleTransform)
        elif isinstance(value, MatrixTransform):
            item.setIcon(icons.matrixTransform)
        elif isinstance(value, TransformGroup):
            item.setIcon(icons.transformGroup)

    def _addChildren(self, root: QStandardItem, value: TransformGroup) -> None:
        for trans in value.children:
            name = trans.__class__.__name__
            item = QStandardItem(name)
            self._setIcon(item, trans)
            self._setUserData(item, trans)
            root.appendRow(item)
            if isinstance(trans, TransformGroup):
                self._addChildren(item, trans)



from ._Geometry import Geometry, GeometryGroup
from ._Curve import CurveGeometry, CurveAlgorithm
from ._Ellipse import EllipseGeometry
from ._Line import LineGeometry, LineAlgorithm, LineClipAlgorithm
from ._LineClipper import LineClipper
from ._Polyline import PolygonGeometry, PolylineGeometry
from ._Rectangle import RectangleGeometry

__all__ = (
    "Geometry",
    "GeometryGroup",
    "LineGeometry",
    "LineAlgorithm",
    "LineClipAlgorithm",
    "EllipseGeometry",
    "CurveGeometry",
    "CurveAlgorithm",
    "PolygonGeometry",
    "PolylineGeometry",
    "RectangleGeometry",
    "LineClipper",
)



from PyQt5.QtCore import QPointF, Qt
from PyQt5.QtGui import QKeyEvent

from ImagingS import Size
from ImagingS.drawing import GeometryDrawing
from ImagingS.geometry import LineGeometry
from ImagingS.Gui import converters

from . import GeometryInteractivity


class LineInteractivity(GeometryInteractivity[LineGeometry]):
    def __init__(self, target: GeometryDrawing, geometry: LineGeometry, size: Size) -> None:
        super().__init__(target, geometry, size)

    def start(self) -> None:
        self._hasStarted = False
        self._isShift = False
        super().start()

    def _setEndPoint(self, point: QPointF) -> None:
        if self._isShift:
            tp = converters.point(point)
            delta = tp - self.geometry.start
            if abs(delta.x) <= abs(delta.y):
                delta.x = 0
            else:
                delta.y = 0
            self.geometry.end = self.geometry.start + delta
        else:
            self.geometry.end = converters.point(point)

    def onMouseRelease(self, point: QPointF) -> None:
        if not self._hasStarted:
            self.geometry.start = converters.point(point)
            self._hasStarted = True
        else:
            self._setEndPoint(point)
            self.update()
            self.end(True)
        super().onMouseRelease(point)

    def onMouseMove(self, point: QPointF) -> None:
        if self._hasStarted:
            self._setEndPoint(point)
            self.update()
        super().onMouseMove(point)

    def onKeyPress(self, key: QKeyEvent) -> None:
        if key.key() == Qt.Key_Shift:
            self._isShift = True
        super().onKeyPress(key)

    def onKeyRelease(self, key: QKeyEvent) -> None:
        if key.key() == Qt.Key_Shift:
            self._isShift = False
        super().onKeyRelease(key)



import importlib
import json
from enum import Enum
from typing import Any, Dict, Tuple

from . import Serializable


def _getTypeName(obj: Any) -> str:
    return f"{obj.__module__}::{obj.__class__.__name__}"


def _splitTypeName(s: str) -> Tuple[str, str]:
    ll = list(s.split("::"))
    return ll[0], ll[1]


class Encoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, Enum):
            return {"__type__": _getTypeName(obj),
                    "__enum__": obj.name}
        elif isinstance(obj, Serializable):
            result = obj.serialize()
            result["__type__"] = _getTypeName(obj)
            return result
        else:
            return super().default(obj)


class Decoder(json.JSONDecoder):
    def __init__(self) -> None:
        json.JSONDecoder.__init__(self, object_hook=Decoder._dict2object)

    @staticmethod
    def _dict2object(dic: Dict) -> Any:
        inst = None
        if "__type__" in dic:
            moduleName, className = _splitTypeName(dic.pop("__type__"))
            module = importlib.import_module(moduleName)
            class_ = getattr(module, className)
            if issubclass(class_, Enum):
                name = dic.pop("__enum__")
                inst = getattr(class_, name)
            elif issubclass(class_, Serializable):
                obj = class_.__call__()
                obj.deserialize(dic)
                inst = obj
            else:
                raise Exception(
                    f"Type '{moduleName}.{className}' is not serializable.")
        else:
            inst = dic
        return inst



from __future__ import annotations

from typing import Optional

from ImagingS import Rect
from ImagingS.brush import Brush, Brushes
from ImagingS.geometry import Geometry

from . import Drawing, Pen, RenderContext


class GeometryDrawing(Drawing):
    def __init__(self, geometry: Optional[Geometry] = None) -> None:
        super().__init__()
        self.geometry = geometry
        self.stroke = Pen()
        self.fill = Brushes.White

    @property
    def stroke(self) -> Pen:
        return self._stroke

    @stroke.setter
    def stroke(self, value: Pen) -> None:
        assert isinstance(value, Pen)
        self._stroke = value

    @property
    def fill(self) -> Brush:
        return self._fill

    @fill.setter
    def fill(self, value: Brush) -> None:
        assert isinstance(value, Brush)
        self._fill = value

    @property
    def geometry(self) -> Optional[Geometry]:
        return self._geometry

    @geometry.setter
    def geometry(self, value: Optional[Geometry]) -> None:
        assert isinstance(value, (type(None), Geometry))
        self._geometry = value

    def render(self, context: RenderContext) -> None:
        if self.geometry is None:
            return
        context.points(self.geometry.fillPoints(), self.fill)
        context.points(self.geometry.strokePoints(self.stroke), self.stroke.brush)

    @property
    def bounds(self) -> Rect:
        if self.geometry is None:
            return Rect()
        else:
            return self.geometry.bounds



from ._Geometry import GeometryInteractivity
from ._Line import LineInteractivity
from ._Polyline import PolylineInteractivity
from ._Curve import CurveInteractivity
from ._Ellipse import EllipseInteractivity
from ._Rectangle import RectangleInteractivity

__all__ = (
    "GeometryInteractivity",
    "LineInteractivity",
    "PolylineInteractivity",
    "CurveInteractivity",
    "EllipseInteractivity",
    "RectangleInteractivity",
)



from abc import ABC, abstractmethod
from typing import Dict

from . import getProperties


class Serializable(ABC):
    @abstractmethod
    def serialize(self) -> Dict:
        pass

    @abstractmethod
    def deserialize(self, data: Dict) -> None:
        pass


class PropertySerializable(Serializable):
    def serialize(self) -> Dict:
        result = {}
        for prop in getProperties(self):
            if prop.canSet:
                result[prop.name] = prop.get()
        return result

    def deserialize(self, data: Dict) -> None:
        for prop in getProperties(self):
            if prop.canSet and prop.name in data:
                value = data[prop.name]
                prop.set(value)
                # if isinstance(value, prop.data_type):
                # else:
                # raise Exception(f"Error type cast: {value.__class__} != {prop.data_type}")



from ._CodePage import CodePage, CodePageState
from ._VisualPage import VisualPage, VisualPageState
from ._DocumentEditor import DocumentEditor, DocumentEditorState
from ._NewDocumentDialog import NewDocumentDialog
from ._MainWindow import MainWindow

__all__ = (
    "NewDocumentDialog",
    "MainWindow",
    "CodePage",
    "VisualPage",
    "CodePageState",
    "VisualPageState",
    "DocumentEditor",
    "DocumentEditorState",
)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/runner/work/ImagingS/ImagingS/src/ImagingS/Gui/uic/CodePage.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_CodePage(object):
    def setupUi(self, CodePage):
        CodePage.setObjectName("CodePage")
        CodePage.resize(400, 300)
        self.horizontalLayout = QtWidgets.QHBoxLayout(CodePage)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.tlbWindow = QtWidgets.QToolBar(CodePage)
        self.tlbWindow.setOrientation(QtCore.Qt.Vertical)
        self.tlbWindow.setObjectName("tlbWindow")
        self.horizontalLayout.addWidget(self.tlbWindow)
        self.widMain = QtWidgets.QWidget(CodePage)
        self.widMain.setObjectName("widMain")
        self.grdCode = QtWidgets.QGridLayout(self.widMain)
        self.grdCode.setContentsMargins(0, 0, 0, 0)
        self.grdCode.setObjectName("grdCode")
        self.swgMain = QtWidgets.QStackedWidget(self.widMain)
        self.swgMain.setObjectName("swgMain")
        self.pagCode = QtWidgets.QWidget()
        self.pagCode.setObjectName("pagCode")
        self.grdCode1 = QtWidgets.QGridLayout(self.pagCode)
        self.grdCode1.setContentsMargins(0, 0, 0, 0)
        self.grdCode1.setObjectName("grdCode1")
        self.tetCode = QtWidgets.QTextEdit(self.pagCode)
        self.tetCode.setObjectName("tetCode")
        self.grdCode1.addWidget(self.tetCode, 0, 0, 1, 1)
        self.swgMain.addWidget(self.pagCode)
        self.pagTree = QtWidgets.QWidget()
        self.pagTree.setObjectName("pagTree")
        self.grdCode2 = QtWidgets.QGridLayout(self.pagTree)
        self.grdCode2.setContentsMargins(0, 0, 0, 0)
        self.grdCode2.setObjectName("grdCode2")
        self.trvTree = QtWidgets.QTreeView(self.pagTree)
        self.trvTree.setObjectName("trvTree")
        self.grdCode2.addWidget(self.trvTree, 0, 0, 1, 1)
        self.swgMain.addWidget(self.pagTree)
        self.grdCode.addWidget(self.swgMain, 0, 0, 1, 1)
        self.horizontalLayout.addWidget(self.widMain)
        self.actRestore = QtWidgets.QAction(CodePage)
        self.actRestore.setObjectName("actRestore")
        self.actCommit = QtWidgets.QAction(CodePage)
        self.actCommit.setObjectName("actCommit")
        self.actSwitch = QtWidgets.QAction(CodePage)
        self.actSwitch.setObjectName("actSwitch")
        self.tlbWindow.addAction(self.actSwitch)
        self.tlbWindow.addAction(self.actRestore)
        self.tlbWindow.addAction(self.actCommit)

        self.retranslateUi(CodePage)
        self.swgMain.setCurrentIndex(1)
        QtCore.QMetaObject.connectSlotsByName(CodePage)

    def retranslateUi(self, CodePage):
        _translate = QtCore.QCoreApplication.translate
        CodePage.setWindowTitle(_translate("CodePage", "Code"))
        self.tlbWindow.setWindowTitle(_translate("CodePage", "Code Tool Box"))
        self.actRestore.setText(_translate("CodePage", "Restore"))
        self.actCommit.setText(_translate("CodePage", "Commit"))
        self.actSwitch.setText(_translate("CodePage", "Switch"))



import glob
import os
import subprocess


def _getParentDir(path: str) -> str:
    return os.path.split(path)[0]


def _getFileNameWithoutExt(path: str) -> str:
    return os.path.split(path)[1].split('.')[0]


def compileUI() -> None:
    curdir = _getParentDir(os.path.realpath(__file__))
    uidir = os.path.join(_getParentDir(curdir), "ui")
    uis = []
    with open(os.path.join(uidir, "__init__.py"), mode="w+") as initf:
        initf.write("# Generated by UI compiler\n\n")
        for src in glob.glob(os.path.join(curdir, "*.ui")):
            name = _getFileNameWithoutExt(src)
            out = os.path.join(uidir, f"_ui_{name}.py")
            print("Compiling", name, "...")
            subprocess.check_call(
                ["python", "-m", "PyQt5.uic.pyuic", src, "-o", out])
            initf.write(f"from ._ui_{name} import Ui_{name} as {name}\n")
            uis.append(name)
        print("Generating UI module...")
        initf.write("\n\n__all__ = (\n")
        for name in uis:
            initf.write(f"    \"{name}\",\n")
        initf.write(")\n")



# Generated by UI compiler

from ._ui_NewDocumentDialog import Ui_NewDocumentDialog as NewDocumentDialog
from ._ui_VisualPage import Ui_VisualPage as VisualPage
from ._ui_DocumentEditor import Ui_DocumentEditor as DocumentEditor
from ._ui_CodePage import Ui_CodePage as CodePage
from ._ui_MainWindow import Ui_MainWindow as MainWindow


__all__ = (
    "NewDocumentDialog",
    "VisualPage",
    "DocumentEditor",
    "CodePage",
    "MainWindow",
)



from __future__ import annotations

from typing import Optional, Tuple

from ImagingS import Point, Rect

from . import LineGeometry, LineClipAlgorithm


def _clipCohenSutherland(start: Point, end: Point, area: Rect) -> Optional[Tuple[Point, Point]]:
    if start == end:
        if start in area:
            return start, end
        else:
            return None
    aori, aver = area.origin, area.vertex()
    LEFT, RIGHT, BOTTOM, TOP = 1, 2, 4, 8

    def encodePoint(p: Point) -> int:
        res = 0
        if p.x < aori.x:
            res |= LEFT
        elif p.x > aver.x:
            res |= RIGHT
        if p.y < aori.y:
            res |= BOTTOM
        elif p.y > aver.y:
            res |= TOP
        return res

    p0, p1 = start.clone(), end.clone()
    c0, c1 = encodePoint(p0), encodePoint(p1)
    while c0 != 0 or c1 != 0:
        if (c0 & c1) != 0:
            return None
        c = c0 if c0 != 0 else c1
        pt = Point()
        if (LEFT & c) != 0:
            pt = Point(aori.x, p0.y + (p1.y - p0.y) * (aori.x - p0.x) / (p1.x - p0.x))
        elif (RIGHT & c) != 0:
            pt = Point(aver.x, p0.y + (p1.y - p0.y) * (aver.x - p0.x) / (p1.x - p0.x))
        elif (BOTTOM & c) != 0:
            pt = Point(p0.x + (p1.x - p0.x) * (aori.y - p0.y) / (p1.y - p0.y), aori.y)
        elif (TOP & c) != 0:
            pt = Point(p0.x + (p1.x - p0.x) * (aver.y - p0.y) / (p1.y - p0.y), aver.y)
        else:
            assert False  # will not full zero, not happend
        if c == c0:
            p0 = pt
            c0 = encodePoint(p0)
        else:
            p1 = pt
            c1 = encodePoint(p1)
    return p0, p1


def _clipLiangBarsky(start: Point, end: Point, area: Rect) -> Optional[Tuple[Point, Point]]:
    if start == end:
        if start in area:
            return start, end
        else:
            return None
    aori, aver = area.origin, area.vertex()
    EPS = 1e-5

    def clipTest(a: float, b: float, t0: float, t1: float) -> Optional[Tuple[float, float]]:
        flag = True
        r0, r1 = t0, t1
        if a < -EPS:
            t = b/a
            if t > t1:
                flag = False
            elif t > t0:
                r0 = t
        elif a > EPS:
            t = b/a
            if t < t0:
                flag = False
            elif t < t1:
                r1 = t
        elif b < 0:
            flag = False
        return (r0, r1) if flag else None

    t0, t1 = 0.0, 1.0
    x0, y0 = start.asTuple()
    x1, y1 = end.asTuple()
    dx, dy = end.x - start.x, end.y - start.y

    res = clipTest(-dx, x0 - aori.x, t0, t1)
    if res is None:
        return None
    t0, t1 = res

    res = clipTest(dx, aver.x - x0, t0, t1)
    if res is None:
        return None
    t0, t1 = res

    res = clipTest(-dy, y0 - aori.y, t0, t1)
    if res is None:
        return None
    t0, t1 = res

    res = clipTest(dy, aver.y - y0, t0, t1)
    if res is None:
        return None
    t0, t1 = res

    if t1 < 1:
        x1, y1 = x0 + t1 * dx, y0 + t1 * dy
    if t0 > 0:
        x0, y0 = x0 + t0 * dx, y0 + t0 * dy

    return Point(x0, y0), Point(x1, y1)


class LineClipper:
    def __init__(self, line: LineGeometry) -> None:
        self.line = line

    def clip(self, area: Rect, algorithm: LineClipAlgorithm) -> Optional[LineGeometry]:
        if area is not None:
            cliped = None
            if algorithm is LineClipAlgorithm.CohenSutherland:
                cliped = _clipCohenSutherland(self.line.start, self.line.end, area)
            elif algorithm is LineClipAlgorithm.LiangBarsky:
                cliped = _clipLiangBarsky(self.line.start, self.line.end, area)
            if cliped is None:
                return None
            else:
                p0, p1 = cliped
                return LineGeometry(p0, p1, self.line.algorithm)



import qtawesome as qta
from PyQt5.QtWidgets import QDialog

import ImagingS.Gui.ui as ui
from ImagingS import Size


class NewDocumentDialog(QDialog, ui.NewDocumentDialog):
    def __init__(self) -> None:
        super().__init__()
        self.setupUi(self)
        self.setupIcon()

    def setupIcon(self) -> None:
        self.setWindowIcon(qta.icon("mdi.file"))

    @property
    def documentSize(self) -> Size:
        return Size(self.sbxWidth.value(), self.sbxHeight.value())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/runner/work/ImagingS/ImagingS/src/ImagingS/Gui/uic/MainWindow.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1028, 717)
        MainWindow.setDockNestingEnabled(True)
        MainWindow.setDockOptions(QtWidgets.QMainWindow.AllowNestedDocks|QtWidgets.QMainWindow.AllowTabbedDocks|QtWidgets.QMainWindow.AnimatedDocks)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        MainWindow.setCentralWidget(self.centralwidget)
        self.mnbWindow = QtWidgets.QMenuBar(MainWindow)
        self.mnbWindow.setGeometry(QtCore.QRect(0, 0, 1028, 26))
        self.mnbWindow.setObjectName("mnbWindow")
        self.mnuFile = QtWidgets.QMenu(self.mnbWindow)
        self.mnuFile.setObjectName("mnuFile")
        self.mnuEdit = QtWidgets.QMenu(self.mnbWindow)
        self.mnuEdit.setObjectName("mnuEdit")
        self.mnuView = QtWidgets.QMenu(self.mnbWindow)
        self.mnuView.setObjectName("mnuView")
        self.mnuBrush = QtWidgets.QMenu(self.mnbWindow)
        self.mnuBrush.setObjectName("mnuBrush")
        self.mnuTool = QtWidgets.QMenu(self.mnbWindow)
        self.mnuTool.setObjectName("mnuTool")
        self.mnuTransform = QtWidgets.QMenu(self.mnbWindow)
        self.mnuTransform.setObjectName("mnuTransform")
        self.mnuDrawing = QtWidgets.QMenu(self.mnbWindow)
        self.mnuDrawing.setObjectName("mnuDrawing")
        MainWindow.setMenuBar(self.mnbWindow)
        self.stbMain = QtWidgets.QStatusBar(MainWindow)
        self.stbMain.setObjectName("stbMain")
        MainWindow.setStatusBar(self.stbMain)
        self.dwgDrawings = QtWidgets.QDockWidget(MainWindow)
        self.dwgDrawings.setAllowedAreas(QtCore.Qt.LeftDockWidgetArea|QtCore.Qt.RightDockWidgetArea)
        self.dwgDrawings.setObjectName("dwgDrawings")
        self.wgDrawings = QtWidgets.QWidget()
        self.wgDrawings.setObjectName("wgDrawings")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.wgDrawings)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.trvDrawings = QtWidgets.QTreeView(self.wgDrawings)
        self.trvDrawings.setContextMenuPolicy(QtCore.Qt.ActionsContextMenu)
        self.trvDrawings.setObjectName("trvDrawings")
        self.verticalLayout.addWidget(self.trvDrawings)
        self.dwgDrawings.setWidget(self.wgDrawings)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(2), self.dwgDrawings)
        self.dwgBrushes = QtWidgets.QDockWidget(MainWindow)
        self.dwgBrushes.setAllowedAreas(QtCore.Qt.LeftDockWidgetArea|QtCore.Qt.RightDockWidgetArea)
        self.dwgBrushes.setObjectName("dwgBrushes")
        self.wgBrushes = QtWidgets.QWidget()
        self.wgBrushes.setObjectName("wgBrushes")
        self.verticalLayout1 = QtWidgets.QVBoxLayout(self.wgBrushes)
        self.verticalLayout1.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout1.setSpacing(0)
        self.verticalLayout1.setObjectName("verticalLayout1")
        self.trvBrushes = QtWidgets.QListView(self.wgBrushes)
        self.trvBrushes.setContextMenuPolicy(QtCore.Qt.ActionsContextMenu)
        self.trvBrushes.setObjectName("trvBrushes")
        self.verticalLayout1.addWidget(self.trvBrushes)
        self.dwgBrushes.setWidget(self.wgBrushes)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(2), self.dwgBrushes)
        self.dwgTransforms = QtWidgets.QDockWidget(MainWindow)
        self.dwgTransforms.setAllowedAreas(QtCore.Qt.LeftDockWidgetArea|QtCore.Qt.RightDockWidgetArea)
        self.dwgTransforms.setObjectName("dwgTransforms")
        self.wgTransforms = QtWidgets.QWidget()
        self.wgTransforms.setObjectName("wgTransforms")
        self.verticalLayout2 = QtWidgets.QVBoxLayout(self.wgTransforms)
        self.verticalLayout2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout2.setSpacing(0)
        self.verticalLayout2.setObjectName("verticalLayout2")
        self.trvTransforms = QtWidgets.QTreeView(self.wgTransforms)
        self.trvTransforms.setContextMenuPolicy(QtCore.Qt.ActionsContextMenu)
        self.trvTransforms.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.trvTransforms.setObjectName("trvTransforms")
        self.verticalLayout2.addWidget(self.trvTransforms)
        self.dwgTransforms.setWidget(self.wgTransforms)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(2), self.dwgTransforms)
        self.dwgProperties = QtWidgets.QDockWidget(MainWindow)
        self.dwgProperties.setAllowedAreas(QtCore.Qt.LeftDockWidgetArea|QtCore.Qt.RightDockWidgetArea)
        self.dwgProperties.setObjectName("dwgProperties")
        self.wgProperties = QtWidgets.QWidget()
        self.wgProperties.setObjectName("wgProperties")
        self.verticalLayout3 = QtWidgets.QVBoxLayout(self.wgProperties)
        self.verticalLayout3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout3.setSpacing(0)
        self.verticalLayout3.setObjectName("verticalLayout3")
        self.trvProperties = QtWidgets.QTreeView(self.wgProperties)
        self.trvProperties.setObjectName("trvProperties")
        self.verticalLayout3.addWidget(self.trvProperties)
        self.dwgProperties.setWidget(self.wgProperties)
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(2), self.dwgProperties)
        self.actNew = QtWidgets.QAction(MainWindow)
        self.actNew.setObjectName("actNew")
        self.actOpen = QtWidgets.QAction(MainWindow)
        self.actOpen.setObjectName("actOpen")
        self.actSave = QtWidgets.QAction(MainWindow)
        self.actSave.setObjectName("actSave")
        self.actSaveAs = QtWidgets.QAction(MainWindow)
        self.actSaveAs.setObjectName("actSaveAs")
        self.actExport = QtWidgets.QAction(MainWindow)
        self.actExport.setObjectName("actExport")
        self.actClose = QtWidgets.QAction(MainWindow)
        self.actClose.setObjectName("actClose")
        self.actAbout = QtWidgets.QAction(MainWindow)
        self.actAbout.setObjectName("actAbout")
        self.actQuit = QtWidgets.QAction(MainWindow)
        self.actQuit.setObjectName("actQuit")
        self.actUndo = QtWidgets.QAction(MainWindow)
        self.actUndo.setObjectName("actUndo")
        self.actRedo = QtWidgets.QAction(MainWindow)
        self.actRedo.setObjectName("actRedo")
        self.actDrawingRemove = QtWidgets.QAction(MainWindow)
        self.actDrawingRemove.setObjectName("actDrawingRemove")
        self.actDrawingClear = QtWidgets.QAction(MainWindow)
        self.actDrawingClear.setObjectName("actDrawingClear")
        self.actBrushSolid = QtWidgets.QAction(MainWindow)
        self.actBrushSolid.setObjectName("actBrushSolid")
        self.actBrushRemove = QtWidgets.QAction(MainWindow)
        self.actBrushRemove.setObjectName("actBrushRemove")
        self.actBrushClear = QtWidgets.QAction(MainWindow)
        self.actBrushClear.setObjectName("actBrushClear")
        self.actTransformRemove = QtWidgets.QAction(MainWindow)
        self.actTransformRemove.setObjectName("actTransformRemove")
        self.actTransformClear = QtWidgets.QAction(MainWindow)
        self.actTransformClear.setObjectName("actTransformClear")
        self.actViewVisual = QtWidgets.QAction(MainWindow)
        self.actViewVisual.setCheckable(True)
        self.actViewVisual.setObjectName("actViewVisual")
        self.actViewCode = QtWidgets.QAction(MainWindow)
        self.actViewCode.setCheckable(True)
        self.actViewCode.setObjectName("actViewCode")
        self.actDrawingsRefresh = QtWidgets.QAction(MainWindow)
        self.actDrawingsRefresh.setObjectName("actDrawingsRefresh")
        self.actTransformsRefresh = QtWidgets.QAction(MainWindow)
        self.actTransformsRefresh.setObjectName("actTransformsRefresh")
        self.actBrushesRefresh = QtWidgets.QAction(MainWindow)
        self.actBrushesRefresh.setObjectName("actBrushesRefresh")
        self.actPropertiesRefresh = QtWidgets.QAction(MainWindow)
        self.actPropertiesRefresh.setObjectName("actPropertiesRefresh")
        self.mnuFile.addAction(self.actNew)
        self.mnuFile.addAction(self.actOpen)
        self.mnuFile.addSeparator()
        self.mnuFile.addAction(self.actSave)
        self.mnuFile.addAction(self.actSaveAs)
        self.mnuFile.addSeparator()
        self.mnuFile.addAction(self.actExport)
        self.mnuFile.addSeparator()
        self.mnuFile.addAction(self.actClose)
        self.mnuFile.addSeparator()
        self.mnuFile.addAction(self.actAbout)
        self.mnuFile.addAction(self.actQuit)
        self.mnuView.addAction(self.actViewVisual)
        self.mnuView.addAction(self.actViewCode)
        self.mnuView.addSeparator()
        self.mnuBrush.addAction(self.actBrushSolid)
        self.mnbWindow.addAction(self.mnuFile.menuAction())
        self.mnbWindow.addAction(self.mnuEdit.menuAction())
        self.mnbWindow.addAction(self.mnuView.menuAction())
        self.mnbWindow.addAction(self.mnuDrawing.menuAction())
        self.mnbWindow.addAction(self.mnuTransform.menuAction())
        self.mnbWindow.addAction(self.mnuBrush.menuAction())
        self.mnbWindow.addAction(self.mnuTool.menuAction())
        self.trvDrawings.addAction(self.actDrawingRemove)
        self.trvDrawings.addAction(self.actDrawingClear)
        self.trvBrushes.addAction(self.actBrushRemove)
        self.trvBrushes.addAction(self.actBrushClear)
        self.trvTransforms.addAction(self.actTransformRemove)
        self.trvTransforms.addAction(self.actTransformClear)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "ImagingS"))
        self.mnuFile.setTitle(_translate("MainWindow", "&File"))
        self.mnuEdit.setTitle(_translate("MainWindow", "&Edit"))
        self.mnuView.setTitle(_translate("MainWindow", "&View"))
        self.mnuBrush.setTitle(_translate("MainWindow", "&Brush"))
        self.mnuTool.setTitle(_translate("MainWindow", "Tool"))
        self.mnuTransform.setTitle(_translate("MainWindow", "&Transform"))
        self.mnuDrawing.setTitle(_translate("MainWindow", "&Drawing"))
        self.dwgDrawings.setWindowTitle(_translate("MainWindow", "Drawings"))
        self.dwgBrushes.setWindowTitle(_translate("MainWindow", "Brushes"))
        self.dwgTransforms.setWindowTitle(_translate("MainWindow", "Transforms"))
        self.dwgProperties.setWindowTitle(_translate("MainWindow", "Properties"))
        self.actNew.setText(_translate("MainWindow", "&New"))
        self.actNew.setShortcut(_translate("MainWindow", "Ctrl+N"))
        self.actOpen.setText(_translate("MainWindow", "&Open"))
        self.actOpen.setShortcut(_translate("MainWindow", "Ctrl+O"))
        self.actSave.setText(_translate("MainWindow", "&Save"))
        self.actSave.setShortcut(_translate("MainWindow", "Ctrl+S"))
        self.actSaveAs.setText(_translate("MainWindow", "Save As"))
        self.actSaveAs.setShortcut(_translate("MainWindow", "Ctrl+Shift+S"))
        self.actExport.setText(_translate("MainWindow", "&Export"))
        self.actClose.setText(_translate("MainWindow", "&Close"))
        self.actClose.setShortcut(_translate("MainWindow", "Ctrl+F4"))
        self.actAbout.setText(_translate("MainWindow", "About"))
        self.actQuit.setText(_translate("MainWindow", "&Quit"))
        self.actUndo.setText(_translate("MainWindow", "Undo"))
        self.actUndo.setShortcut(_translate("MainWindow", "Ctrl+Z"))
        self.actRedo.setText(_translate("MainWindow", "Redo"))
        self.actRedo.setShortcut(_translate("MainWindow", "Ctrl+Y"))
        self.actDrawingRemove.setText(_translate("MainWindow", "Remove"))
        self.actDrawingClear.setText(_translate("MainWindow", "Clear"))
        self.actBrushSolid.setText(_translate("MainWindow", "Solid"))
        self.actBrushRemove.setText(_translate("MainWindow", "Remove"))
        self.actBrushClear.setText(_translate("MainWindow", "Clear"))
        self.actTransformRemove.setText(_translate("MainWindow", "Remove"))
        self.actTransformClear.setText(_translate("MainWindow", "Clear"))
        self.actViewVisual.setText(_translate("MainWindow", "Visual"))
        self.actViewVisual.setShortcut(_translate("MainWindow", "Ctrl+Shift+V"))
        self.actViewCode.setText(_translate("MainWindow", "Code"))
        self.actViewCode.setShortcut(_translate("MainWindow", "Ctrl+Shift+C"))
        self.actDrawingsRefresh.setText(_translate("MainWindow", "Refresh"))
        self.actTransformsRefresh.setText(_translate("MainWindow", "Refresh"))
        self.actBrushesRefresh.setText(_translate("MainWindow", "Refresh"))
        self.actPropertiesRefresh.setText(_translate("MainWindow", "Refresh"))



import math
import os
from typing import List, cast

from PIL import Image
import numpy as np

from ImagingS import Color, Point, Rect, Size
from ImagingS.brush import Brushes, SolidBrush
from ImagingS.document import Document, DocumentFormat
from ImagingS.drawing import GeometryDrawing, NumpyArrayRenderContext, Pen
from ImagingS.geometry import (CurveAlgorithm, CurveGeometry, EllipseGeometry,
                               LineAlgorithm, LineClipAlgorithm, LineGeometry,
                               LineClipper, PolygonGeometry)
from ImagingS.transform import (RotateTransform, ScaleTransform, Transform,
                                TransformGroup, TranslateTransform)


def _append_transform(drawing: GeometryDrawing, transform: Transform) -> None:
    assert not isinstance(transform, TransformGroup)
    geometry = drawing.geometry
    if geometry.transform is None:
        geometry.transform = transform
    elif isinstance(geometry.transform, TransformGroup):
        geometry.transform.children.append(transform)
    else:
        old = geometry.transform
        new = TransformGroup()
        new.children.append(old)
        new.children.append(transform)
        geometry.transform = new


class BuiltinInstruction:
    def __init__(self, doc: Document, output_dir: str) -> None:
        self.doc = doc
        self.brush = Brushes.Black
        self.output_dir = output_dir

    def resetCanvas(self, argv: List[str]) -> None:
        self.doc.size = Size(int(argv[0]), int(argv[1]))
        self.doc.drawings.children.clear()

    def saveDoc(self, argv: List[str]) -> None:
        fileName = os.path.join(self.output_dir, f"{argv[0]}.isd.json")

        with open(fileName, "w+") as f:
            self.doc.save(f, DocumentFormat.RAW)

        fileName = os.path.join(self.output_dir, f"{argv[0]}.isd")

        with open(fileName, "wb") as f:
            self.doc.save(f)

    def saveCanvas(self, argv: List[str]) -> None:
        fileName = os.path.join(self.output_dir, f"{argv[0]}.bmp")

        context = NumpyArrayRenderContext(
            NumpyArrayRenderContext.create_array(self.doc.size))

        self.doc.drawings.render(context)
        Image.fromarray(np.flipud(context.array)).save(fileName, "bmp")

    def setColor(self, argv: List[str]) -> None:
        self.brush = SolidBrush(
            Color(int(argv[0]), int(argv[1]), int(argv[2])))

    def drawLine(self, argv: List[str]) -> None:
        geometry = LineGeometry(Point(int(argv[1]), int(argv[2])), Point(
            int(argv[3]), int(argv[4])), LineAlgorithm.Dda if argv[5] == "DDA" else LineAlgorithm.Bresenham)
        drawing = GeometryDrawing(geometry)
        drawing.id = argv[0]
        drawing.stroke = Pen(self.brush)
        self.doc.drawings.children.append(drawing)

    def drawPolygon(self, argv: List[str]) -> None:
        vers: List[Point] = []
        ps = argv[1:-1]
        i = 0
        while i < len(ps):
            vers.append(Point(int(ps[i]), int(ps[i+1])))
            i += 2
        geometry = PolygonGeometry(
            vers, LineAlgorithm.Dda if argv[-1] == "DDA" else LineAlgorithm.Bresenham)
        drawing = GeometryDrawing(geometry)
        drawing.id = argv[0]
        drawing.stroke = Pen(self.brush)
        self.doc.drawings.children.append(drawing)

    def drawEllipse(self, argv: List[str]) -> None:
        lt = Point(int(argv[1]), int(argv[2]))
        rb = Point(int(argv[3]), int(argv[4]))
        geometry = EllipseGeometry.fromRect(Rect.fromPoints(lt, rb))
        drawing = GeometryDrawing(geometry)
        drawing.id = argv[0]
        drawing.stroke = Pen(self.brush)
        self.doc.drawings.children.append(drawing)

    def drawCurve(self, argv: List[str]) -> None:
        vers: List[Point] = []
        ps = argv[1:-1]
        i = 0
        while i < len(ps):
            vers.append(Point(int(ps[i]), int(ps[i+1])))
            i += 2
        geometry = CurveGeometry(
            vers, CurveAlgorithm.Bezier if argv[-1] == "Bezier" else CurveAlgorithm.BSpline)
        drawing = GeometryDrawing(geometry)
        drawing.id = argv[0]
        drawing.stroke = Pen(self.brush)
        self.doc.drawings.children.append(drawing)

    def translate(self, argv: List[str]) -> None:
        drawing = self.doc.drawings.children[argv[0]]
        assert isinstance(drawing, GeometryDrawing)
        _append_transform(drawing, TranslateTransform(
            Point(int(argv[1]), int(argv[2]))))

    def rotate(self, argv: List[str]) -> None:
        drawing = self.doc.drawings.children[argv[0]]
        assert isinstance(drawing, GeometryDrawing)
        _append_transform(drawing, RotateTransform(
            Point(int(argv[1]), int(argv[2])), -int(argv[3]) / 360 * 2 * math.pi))

    def scale(self, argv: List[str]) -> None:
        drawing = self.doc.drawings.children[argv[0]]
        assert isinstance(drawing, GeometryDrawing)
        _append_transform(drawing, ScaleTransform(
            Point(int(argv[1]), int(argv[2])), (float(argv[3]), float(argv[3]))))

    def clip(self, argv: List[str]) -> None:
        drawing = self.doc.drawings.children[argv[0]]
        assert isinstance(drawing, GeometryDrawing)
        assert isinstance(drawing.geometry, LineGeometry)
        lt = Point(int(argv[1]), int(argv[2]))
        rb = Point(int(argv[3]), int(argv[4]))

        area = Rect.fromPoints(lt, rb)
        clipAlgorithm = LineClipAlgorithm.CohenSutherland if argv[
            5] == "Cohen-Sutherland" else LineClipAlgorithm.LiangBarsky

        geo = drawing.geometry
        clipper = LineClipper(cast(LineGeometry, geo.transformed()))
        clipped = clipper.clip(area, clipAlgorithm)

        if clipped is None:
            parent = drawing.parent()
            if parent is None:
                return
            del parent[drawing.id]
        else:
            geo.start = clipped.start
            geo.end = clipped.end
            geo.transform = None  # clear transform

    def execute(self, ins: str) -> None:
        ins = ins.strip()
        if not ins:
            return
        items = ins.split(" ")
        m = getattr(self, items[0])
        m.__call__(items[1:])



from ._BrushModel import BrushModel
from ._DrawingModel import DrawingModel
from ._PropertyModel import PropertyModel
from ._TransformModel import TransformModel
from ._CommitCommand import CommitCommand

__all__ = (
    "BrushModel",
    "DrawingModel",
    "PropertyModel",
    "TransformModel",
    "CommitCommand"
)



from __future__ import annotations

from datetime import datetime
from io import BytesIO
from typing import List, Optional

from . import Document, DocumentFormat


class Commit:
    def __init__(self, time: datetime, data: bytes, message: str, parent: Optional[Commit]) -> None:
        self.time = time
        self.message = message
        self.data = data
        self.parent = parent


class VersionController:
    def __init__(self) -> None:
        self.commits: List[Commit] = []
        self.head: Optional[Commit] = None

    def commit(self, doc: Document, message: str) -> Commit:
        io = BytesIO()
        doc.save(io, DocumentFormat.ISD)
        data = io.getvalue()
        cmt = Commit(datetime.now(), data, message, self.head)
        self.commits.append(cmt)
        self.head = cmt
        return cmt

    def getDocument(self, commit: Commit) -> Document:
        io = BytesIO(commit.data)
        return Document.load(io, DocumentFormat.ISD)

    def clear(self):
        self.commits.clear()
        self.head = None



from typing import Any

from PyQt5.QtCore import QModelIndex, Qt, pyqtSignal
from PyQt5.QtGui import QIcon, QStandardItem, QStandardItemModel

from ImagingS.drawing import Drawing, DrawingGroup, GeometryDrawing
from ImagingS.geometry import (CurveGeometry, EllipseGeometry, LineGeometry,
                               PolygonGeometry, PolylineGeometry,
                               RectangleGeometry)
from ImagingS.Gui import icons


class DrawingModel(QStandardItemModel):
    ID, TYPE = range(2)
    changed = pyqtSignal()

    def __init__(self, parent) -> None:
        super().__init__(0, 1, parent)
        self.setHeaderData(DrawingModel.ID, Qt.Horizontal, "Id")
        self.setHeaderData(DrawingModel.TYPE, Qt.Horizontal, "Type")
        self.obj = None

    def fresh(self, obj: Drawing = None) -> None:
        self.removeRows(0, self.rowCount())
        self.obj = obj
        if obj is None:
            return

        type = obj.__class__.__name__
        item = QStandardItem()
        item.setEditable(False)
        index = self.indexFromItem(item).row()
        typeItem = QStandardItem(type)
        typeItem.setEditable(False)
        self.setItem(index, self.TYPE, typeItem)
        self._setUserData(item, obj)
        self.appendRow(item)

        if isinstance(obj, DrawingGroup):
            self._addChildren(item, obj)

    def data(self, index: QModelIndex, role: int) -> Any:
        if index.column() == self.ID:
            drawing = self.getUserData(index)
            if role == Qt.DisplayRole or role == Qt.EditRole:
                if drawing == self.obj:  # Root
                    return "Document"
                else:
                    return drawing.id
            elif role == Qt.DecorationRole:
                return self._getIcon(drawing)
        return super().data(index, role)

    def setData(self, index: QModelIndex, value: Any, role: int) -> bool:
        if role == Qt.EditRole:
            if index.column() == self.ID:
                drawing = self.getUserData(index)
                try:
                    parent = drawing.parent()
                    if parent is not None:
                        parent.setItemId(drawing.id, str(value))
                    else:
                        drawing.id = str(value)
                except Exception:
                    return False
                self.dataChanged.emit(index, index, [role])
                self.changed.emit()
                return True
        return super().setData(index, value, role)

    def getUserData(self, index: QModelIndex) -> Drawing:
        item = self.itemFromIndex(index)
        return item.data(Qt.UserRole)

    def _setUserData(self, item: QStandardItem, value: Drawing) -> None:
        item.setData(value, Qt.UserRole)

    def _getIcon(self, value: Drawing) -> QIcon:
        if isinstance(value, GeometryDrawing):
            if isinstance(value.geometry, LineGeometry):
                return icons.lineGeometry
            elif isinstance(value.geometry, CurveGeometry):
                return icons.curveGeometry
            elif isinstance(value.geometry, EllipseGeometry):
                return icons.ellipseGeometry
            elif isinstance(value.geometry, RectangleGeometry):
                return icons.rectangleGeometry
            elif isinstance(value.geometry, PolygonGeometry):
                return icons.polygonGeometry
            elif isinstance(value.geometry, PolylineGeometry):
                return icons.polylineGeometry
        return icons.drawing

    def _addChildren(self, root: QStandardItem, value: DrawingGroup) -> None:
        for drawing in value.children:
            type = drawing.__class__.__name__
            item = QStandardItem(drawing.id)
            self._setUserData(item, drawing)
            root.appendRow(item)
            index = self.indexFromItem(item).row()

            typeItem = QStandardItem(type)
            typeItem.setEditable(False)
            root.setChild(index, self.TYPE, typeItem)

            if isinstance(drawing, DrawingGroup):
                self._addChildren(item, drawing)



from __future__ import annotations

from typing import Callable

from PyQt5.QtWidgets import QUndoCommand

from ImagingS.document import Commit, Document, VersionController


class CommitCommand(QUndoCommand):
    def __init__(self, commit: Commit, controller: VersionController, callback: Callable[[Document], None], firstredo: bool = False) -> None:
        super().__init__()
        self.commit = commit
        self.controller = controller
        self.callback = callback
        if not firstredo:
            self.firstredo = False
        self.setText(self.commit.message)

    def undo(self):
        assert self.commit.parent is not None
        self.controller.head = self.commit.parent
        self.callback(self.controller.getDocument(self.commit.parent))

    def redo(self):
        if hasattr(self, "firstredo"):
            del self.firstredo
            return
        self.controller.head = self.commit
        self.callback(self.controller.getDocument(self.commit))



from typing import Union

from PyQt5.QtCore import QPoint, QPointF, QRect, QRectF, QSize, QSizeF
from PyQt5.QtGui import QColor

from ImagingS import Color, Point, Rect, Size


def qpoint(o: Point) -> QPointF:
    return QPointF(o.x, o.y)


def qsize(o: Size) -> QSizeF:
    return QSizeF(o.width, o.height)


def qrect(o: Rect) -> QRectF:
    return QRectF(qpoint(o.origin), qsize(o.size))


def qcolor(o: Color) -> QColor:
    return QColor(o.r, o.g, o.b)


def point(o: Union[QPointF, QPoint]) -> Point:
    return Point(o.x(), o.y())


def size(o: Union[QSizeF, QSize]) -> Size:
    return Size(o.width(), o.height())


def rect(o: Union[QRectF, QRect]) -> Rect:
    return Rect(point(o.topLeft()), size(o.size()))


def color(o: QColor) -> Color:
    return Color(o.red(), o.green(), o.blue())



from abc import ABC, abstractmethod
from typing import Callable, Iterable

import numpy as np

from ImagingS import Color, Point, Rect, Size
from ImagingS.brush import Brush


class RenderContext(ABC):
    @abstractmethod
    def _point(self, position: Point, color: Color) -> None: pass

    @abstractmethod
    def bounds(self) -> Rect: pass

    def point(self, position: Point, color: Color) -> None:
        if position in self.bounds():
            self._point(position, color)

    def points(self, positions: Iterable[Point], brush: Brush) -> None:
        bounds = self.bounds()
        for p in positions:
            self.point(p, brush.colorAt(p, bounds))


class NumpyArrayRenderContext(RenderContext):
    @staticmethod
    def create_array(size: Size) -> np.ndarray:
        result = np.zeros([int(size.height), int(size.width), 3], np.uint8)
        result.fill(255)
        return result

    def __init__(self, array: np.ndarray) -> None:
        self.array = array

    @property
    def array(self) -> np.ndarray:
        super().__init__()
        return self._array

    @array.setter
    def array(self, value: np.ndarray) -> None:
        assert isinstance(value, np.ndarray)
        assert value.dtype == np.uint8
        assert len(value.shape) == 3
        assert value.shape[2] == 3
        self._array = value
        self._bounds = Rect(
            Point(), Size(value.shape[1], value.shape[0]))

    def _point(self, position: Point, color: Color) -> None:
        x, y = map(int, position.asTuple())
        self.array[y, x, 0] = color.r
        self.array[y, x, 1] = color.g
        self.array[y, x, 2] = color.b

    def bounds(self) -> Rect:
        return self._bounds


class ProxyRenderContext(RenderContext):
    def __init__(self, fpoint: Callable[[Point, Color], None], fbounds: Callable[[], Rect]) -> None:
        super().__init__()
        self._fpoint = fpoint
        self._fbounds = fbounds

    def _point(self, position: Point, color: Color) -> None:
        self._fpoint(position, color)

    def bounds(self) -> Rect:
        return self._fbounds()


class ClipRenderContext(ProxyRenderContext):
    def __init__(self, bounds: Rect, innerContext: RenderContext) -> None:
        super().__init__(innerContext.point, lambda: bounds)



from enum import Enum, unique
from io import StringIO

import qtawesome as qta
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QWidget

import ImagingS.Gui.ui as ui
from ImagingS.document import Document, DocumentFormat
from ImagingS.Gui import icons
from ImagingS.Gui.models import PropertyModel


@unique
class CodePageState(Enum):
    Disable = 0,
    Normal = 1,
    Changed = 2


def _clonedoc(doc: Document) -> Document:
    io = StringIO()
    doc.save(io, DocumentFormat.RAW)
    io = StringIO(io.getvalue())
    return Document.load(io, DocumentFormat.RAW)


class CodePage(QWidget, ui.CodePage):
    documentCommitted = pyqtSignal(Document, str)
    messaged = pyqtSignal(str)
    stateChanged = pyqtSignal()

    def __init__(self) -> None:
        super().__init__()
        self.setupUi(self)
        self.setupIcon()

        self.modelCode = PropertyModel(self)
        self.modelCode.onlyEditable = True
        self.trvTree.setModel(self.modelCode)
        self.modelCode.changed.connect(self.modelCode_changed)
        self.tetCode.textChanged.connect(self.tetCode_textChanged)

        self.actRestore.triggered.connect(self.actRestore_triggered)
        self.actCommit.triggered.connect(self.actCommit_triggered)
        self.actSwitch.triggered.connect(self.actSwitch_triggered)

        self.swgMain.setCurrentIndex(0)

        self.setEnabled(False)
        self._state = CodePageState.Disable

    def setupIcon(self) -> None:
        self.actRestore.setIcon(qta.icon("mdi.restore"))
        self.actCommit.setIcon(qta.icon("mdi.check"))
        self.actSwitch.setIcon(icons.fileTree)
        self.setWindowIcon(icons.code)

    def switchCode(self, force: bool = False) -> None:
        if self.swgMain.currentIndex() == 0 and not force:
            return

        io = StringIO()
        self._curdocument.save(io, DocumentFormat.RAW)
        self.tetCode.setText(io.getvalue())
        self.actSwitch.setIcon(icons.fileTree)
        self.swgMain.setCurrentIndex(0)

    def switchTree(self, force: bool = False) -> None:
        if self.swgMain.currentIndex() == 1 and not force:
            return

        io = StringIO(self.tetCode.toPlainText())
        try:
            tdoc = Document.load(io, DocumentFormat.RAW)
        except Exception:
            self.messaged.emit("Some syntax errors in code")
            return

        self._curdocument = tdoc
        self.modelCode.fresh(self._curdocument)
        width = round(self.trvTree.size().width() / 2)
        if width > 0:
            self.trvTree.setColumnWidth(0, width)
            self.trvTree.setColumnWidth(1, width)
        self.trvTree.expandAll()

        self.actSwitch.setIcon(icons.code)
        self.swgMain.setCurrentIndex(1)

    def _setState(self, value: CodePageState) -> None:
        self._state = value
        if self.state is CodePageState.Changed:
            self.actRestore.setEnabled(True)
            self.actCommit.setEnabled(True)
        else:
            self.actRestore.setEnabled(False)
            self.actCommit.setEnabled(False)
        self.stateChanged.emit()

    def fresh(self) -> None:
        if self.state is not CodePageState.Normal:
            return

        if self.swgMain.currentIndex() == 0:
            self.switchCode(True)
        else:
            self.switchTree(True)
        self._setState(CodePageState.Normal)

    def enable(self, doc: Document) -> None:
        assert self.state is CodePageState.Disable
        self.setEnabled(True)
        self._document = doc
        self._curdocument = _clonedoc(self._document)
        self.switchCode(True)
        self._setState(CodePageState.Normal)

    def disable(self) -> None:
        assert self.state is not CodePageState.Disable
        if hasattr(self, "_document"):
            del self._document
        self.tetCode.setText("")
        self.modelCode.fresh()
        self.setEnabled(False)
        self._setState(CodePageState.Disable)

    @property
    def state(self) -> CodePageState:
        return self._state

    def modelCode_changed(self) -> None:
        self._setState(CodePageState.Changed)

    def tetCode_textChanged(self) -> None:
        self._setState(CodePageState.Changed)

    def actRestore_triggered(self) -> None:
        self._curdocument = _clonedoc(self._document)
        self.switchCode(True)
        self._setState(CodePageState.Normal)

    def actCommit_triggered(self) -> None:
        if self.swgMain.currentIndex() == 0 and self.state is CodePageState.Changed:
            io = StringIO(self.tetCode.toPlainText())
            try:
                tdoc = Document.load(io, DocumentFormat.RAW)
            except Exception:
                self.messaged.emit("Some syntax errors in code")
                return

            self._curdocument = tdoc

        self.modelCode.fresh(self._curdocument)
        self._document = self._curdocument
        self.actRestore.trigger()
        self.documentCommitted.emit(self._document, "Change Code")

    def actSwitch_triggered(self) -> None:
        if self.swgMain.currentIndex() == 0:
            self.switchTree()
        else:
            self.switchCode()



from typing import Dict, Optional

from PyQt5.QtCore import QPointF, QRectF, QSizeF, Qt, pyqtSignal
from PyQt5.QtGui import QColor, QKeyEvent, QMouseEvent, QPainterPath, QPen
from PyQt5.QtWidgets import (QAction, QGraphicsPathItem, QGraphicsRectItem,
                             QGraphicsScene, QGraphicsView)

from ImagingS.drawing import Drawing
from ImagingS.Gui import icons
from ImagingS.Gui.interactivity import Interactivity

from . import DrawingItem


class Canvas(QGraphicsView):
    mouseMoved = pyqtSignal(QPointF)
    mouseReleased = pyqtSignal(QPointF)

    def __init__(self, parent):
        scene = QGraphicsScene(parent)
        super().__init__(scene, parent)
        self.setupActions()
        self.setContextMenuPolicy(Qt.ActionsContextMenu)

        self.interactivity = None

        self.items: Dict[str, DrawingItem] = {}
        self._active_item: Optional[DrawingItem] = None

        self._sceneBorder = QGraphicsRectItem()
        self._sceneGrid = QGraphicsPathItem()
        self._sceneGrid.setVisible(False)
        self._sceneGrid.setPen(QPen(QColor("lightgray")))
        self.scene().addItem(self._sceneBorder)
        self.scene().addItem(self._sceneGrid)

        self.setMouseTracking(True)

    def setupActions(self):
        self.actRerender = QAction(self)
        self.actRerender.setObjectName("actactRerender")
        self.actRerender.triggered.connect(self.actRerender_triggered)
        self.actRerender.setIcon(icons.refresh)
        self.actRerender.setText("Rerender")
        self.actRerender.setShortcut("F5")
        self.addAction(self.actRerender)

        self.actGrid = QAction(self)
        self.actGrid.setObjectName("actGrid")
        self.actGrid.triggered.connect(self.actGrid_triggered)
        self.actGrid.setIcon(icons.grid)
        self.actGrid.setCheckable(True)
        self.actGrid.setChecked(False)
        self.actGrid.setText("Gird")
        self.actGrid.setShortcut("F4")
        self.addAction(self.actGrid)

    @property
    def interactivity(self) -> Optional[Interactivity]:
        return self._interactivity

    @interactivity.setter
    def interactivity(self, value: Optional[Interactivity]) -> None:
        self._interactivity = value

    def rerender(self) -> None:
        self.updateScene([self.sceneRect()])

    def resize(self, size: QSizeF):
        self.scene().setSceneRect(QRectF(QPointF(), size))
        self._sceneBorder.setRect(self.scene().sceneRect())

        gridPath = QPainterPath()
        D = 80
        for x in range(0, round(size.width()), D):
            gridPath.moveTo(QPointF(x, 0))
            gridPath.lineTo(QPointF(x, size.height()))
        for y in range(0, round(size.height()), D):
            gridPath.moveTo(QPointF(0, y))
            gridPath.lineTo(QPointF(size.width(), y))
        self._sceneGrid.setPath(gridPath)

        draws = self.items.values()

        self.clear()
        for item in draws:
            self.add(item.drawing)

    def add(self, drawing: Drawing) -> None:
        item = DrawingItem(drawing, self.scene().sceneRect().size())
        self.items[drawing.id] = item
        self.scene().addItem(item)
        self.rerender()

    def remove(self, name: str) -> None:
        if name not in self.items:
            return
        item = self.items[name]
        del self.items[name]
        self.scene().removeItem(item)
        self.rerender()

    def select(self, name: Optional[str] = None) -> None:
        if self._active_item is not None:
            self._active_item.deactivate()
        if name is None or name not in self.items:
            self._active_item = None
        else:
            item = self.items[name]
            item.activate()
            self._active_item = item
        self.rerender()

    def clear(self) -> None:
        for draw in self.items.values():
            self.scene().removeItem(draw)
        self.items.clear()
        self.rerender()

    def actRerender_triggered(self) -> None:
        self.rerender()

    def actGrid_triggered(self) -> None:
        if self.actGrid.isChecked():
            self._sceneGrid.setVisible(True)
        else:
            self._sceneGrid.setVisible(False)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        pos = self.mapToScene(event.localPos().toPoint())
        if self.interactivity is not None:
            inter = self.interactivity
            inter.onMousePress(pos)
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        pos = self.mapToScene(event.localPos().toPoint())
        if self.interactivity is not None:
            inter = self.interactivity
            inter.onMouseMove(pos)
        self.mouseMoved.emit(pos)
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        pos = self.mapToScene(event.localPos().toPoint())
        if self.interactivity is not None:
            inter = self.interactivity
            inter.onMouseRelease(pos)
        self.mouseReleased.emit(pos)
        super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        pos = self.mapToScene(event.localPos().toPoint())
        if self.interactivity is not None:
            inter = self.interactivity
            inter.onMouseDoubleClick(pos)
        super().mouseDoubleClickEvent(event)

    def keyPressEvent(self, event: QKeyEvent) -> None:
        if self.interactivity is not None:
            inter = self.interactivity
            inter.onKeyPress(event)
        super().keyPressEvent(event)

    def keyReleaseEvent(self, event: QKeyEvent) -> None:
        if self.interactivity is not None:
            inter = self.interactivity
            inter.onKeyRelease(event)
        super().keyReleaseEvent(event)



from __future__ import annotations

from typing import Iterable, List, Optional, cast

from ImagingS import Point, Rect, RectMeasurer
from ImagingS.drawing import Pen

from . import Geometry, LineAlgorithm, LineGeometry


class PolylineGeometry(Geometry):
    S_Vertexes = "vertexes"

    def __init__(self, vertexes: Optional[List[Point]] = None, algorithm: Optional[LineAlgorithm] = None) -> None:
        super().__init__()
        self.vertexes = vertexes if vertexes else []
        self.algorithm = algorithm if algorithm else LineAlgorithm.Dda

    @property
    def algorithm(self) -> LineAlgorithm:
        return self._algorithm

    @algorithm.setter
    def algorithm(self, value: LineAlgorithm) -> None:
        assert isinstance(value, LineAlgorithm)
        self._algorithm = value

    @property
    def vertexes(self) -> List[Point]:
        return self._vertexes

    @vertexes.setter
    def vertexes(self, value: List[Point]) -> None:
        assert isinstance(value, list)
        self._vertexes = value
        self.refreshBounds()

    def transformed(self) -> Geometry:
        if self.transform is None:
            return self
        else:
            vertexes = list(map(self.transform.transform, self.vertexes))
            return PolylineGeometry(vertexes, self.algorithm)

    def strokePoints(self, pen: Pen) -> Iterable[Point]:
        target = self
        if self.transform is not None:
            target = cast(PolylineGeometry, self.transformed())
        cnt = len(target.vertexes)
        if cnt == 0:
            return
        for i in range(cnt - 1):
            ln = LineGeometry(target.vertexes[i],
                              target.vertexes[i+1], target.algorithm)
            for point in ln.strokePoints(pen):
                yield point

    def fillPoints(self) -> Iterable[Point]:
        return []

    def _calculateBounds(self) -> Rect:
        target = self
        if self.transform is not None:
            target = cast(PolylineGeometry, self.transformed())
        measure = RectMeasurer()
        measure.append(target.vertexes)
        return measure.result()


class PolygonGeometry(PolylineGeometry):
    def __init__(self, vertexes: Optional[List[Point]] = None, algorithm: Optional[LineAlgorithm] = None) -> None:
        super().__init__(vertexes, algorithm)

    def transformed(self) -> Geometry:
        if self.transform is None:
            return self
        else:
            vertexes = list(map(self.transform.transform, self.vertexes))
            return PolygonGeometry(vertexes, self.algorithm)

    def strokePoints(self, pen: Pen) -> Iterable[Point]:
        target = self
        if self.transform is not None:
            target = cast(PolygonGeometry, self.transformed())
        cnt = len(target.vertexes)
        if cnt > 0:
            for point in super(PolygonGeometry, target).strokePoints(pen):
                yield point
            ln = LineGeometry(target.vertexes[-1],
                              target.vertexes[0], target.algorithm)
            for point in ln.strokePoints(pen):
                yield point



from typing import Union

from PyQt5.QtCore import QPointF, Qt
from PyQt5.QtGui import QKeyEvent, QPainterPath

from ImagingS.drawing import DrawingGroup
from ImagingS.geometry import Geometry
from ImagingS.Gui import converters
from ImagingS.transform import TranslateTransform

from . import TransformInteractivity


class TranslateTransformInteractivity(TransformInteractivity[TranslateTransform]):
    def __init__(self, target: Union[DrawingGroup, Geometry], transform: TranslateTransform) -> None:
        super().__init__(target, transform)

    def start(self) -> None:
        self._hasStarted = False
        self._isShift = False
        super().start()

    def _updateData(self) -> None:
        self.transform.delta = self._end - self._start

        path = QPainterPath()
        s = converters.qpoint(self._start)
        t = converters.qpoint(self._end)
        path.addEllipse(s, 3, 3)
        path.moveTo(s)
        path.lineTo(t)
        self._viewItem.setPath(path)

    def _setEndPoint(self, point: QPointF) -> None:
        if self._isShift:
            tp = converters.point(point)
            delta = tp - self._start
            if abs(delta.x) <= abs(delta.y):
                delta.x = 0
            else:
                delta.y = 0
            self._end = self._start + delta
        else:
            self._end = converters.point(point)

    def onMouseRelease(self, point: QPointF) -> None:
        if not self._hasStarted:
            self._start = converters.point(point)
            self._hasStarted = True
        else:
            self._setEndPoint(point)
            self._updateData()
            self.end(True)
        super().onMouseRelease(point)

    def onMouseMove(self, point: QPointF) -> None:
        if self._hasStarted:
            self._setEndPoint(point)
            self._updateData()
            self.update()
        super().onMouseMove(point)

    def onKeyPress(self, key: QKeyEvent) -> None:
        if key.key() == Qt.Key_Shift:
            self._isShift = True
        super().onKeyPress(key)

    def onKeyRelease(self, key: QKeyEvent) -> None:
        if key.key() == Qt.Key_Shift:
            self._isShift = False
        super().onKeyRelease(key)



from __future__ import annotations

from typing import Optional

from ImagingS import Color, Colors, Point, Rect

from . import Brush


class SolidBrush(Brush):
    def __init__(self, color: Optional[Color] = None) -> None:
        super().__init__()
        self.color = color if color else Colors.Black

    @property
    def color(self) -> Color:
        return self._color

    @color.setter
    def color(self, value: Color) -> None:
        self._color = value

    def colorAt(self, position: Point, rect: Rect) -> Color:
        return self.color

    def __repr__(self) -> str:
        return f"SolidBrush({self.color})"



from typing import Generic, Optional, TypeVar

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QKeyEvent
from PyQt5.QtWidgets import QGraphicsItem

from ImagingS import Size
from ImagingS.drawing import GeometryDrawing
from ImagingS.geometry import Geometry
from ImagingS.Gui import converters
from ImagingS.Gui.graphic import DrawingItem

from .. import Interactivity

_T = TypeVar("_T")


class GeometryInteractivity(Interactivity, Generic[_T]):
    def __init__(self, target: GeometryDrawing, geometry: _T, size: Size) -> None:
        super().__init__()
        assert isinstance(geometry, Geometry)
        self._target = target
        self._viewItem = DrawingItem(self._target, converters.qsize(size))
        self._geometry = geometry

    @property
    def viewItem(self) -> Optional[QGraphicsItem]:
        return self._viewItem

    @property
    def target(self) -> GeometryDrawing:
        return self._target

    @property
    def geometry(self) -> _T:
        return self._geometry

    def start(self) -> None:
        self.target.geometry = self.geometry
        super().start()

    def end(self, success: bool) -> None:
        if not success:
            self.target.geometry = None
        super().end(success)

    def update(self) -> None:
        self.geometry.refreshBounds()
        super().update()

    def onKeyPress(self, key: QKeyEvent) -> None:
        if key.key() == Qt.Key_Escape:
            self.end(False)
        super().onKeyPress(key)



from typing import Optional

from PyQt5.QtCore import QRectF, QSizeF
from PyQt5.QtGui import QColor, QPainter
from PyQt5.QtWidgets import QGraphicsItem, QStyleOptionGraphicsItem, QWidget

from ImagingS import Point, Rect
from ImagingS.drawing import Drawing, GeometryDrawing
from ImagingS.geometry import Geometry
from ImagingS.Gui import converters

from . import PainterRenderContext


class DrawingItem(QGraphicsItem):
    def __init__(self, drawing: Drawing, size: QSizeF, parent: Optional[QGraphicsItem] = None):
        super().__init__(parent)
        self._drawing = drawing
        self._size = size
        self._is_active = False

    @property
    def drawing(self) -> Drawing:
        return self._drawing

    def activate(self) -> None:
        if isinstance(self.drawing, GeometryDrawing):
            self.drawing.geometry.refreshBounds()
        self._is_active = True

    def deactivate(self) -> None:
        self._is_active = False

    def paint(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget: Optional[QWidget] = None) -> None:
        context = PainterRenderContext(
            painter, Rect(Point(), converters.size(self._size)))
        self.drawing.render(context)
        if self._is_active:
            painter.setPen(QColor(255, 0, 0))
            rect = self.drawing.bounds
            painter.drawRect(converters.qrect(rect))

    def boundingRect(self) -> QRectF:  # must be efficient
        # to fix prepareGeometryChange bug
        return QRectF(0, 0, self._size.width(), self._size.height())
        # self.prepareGeometryChange()  # important!!!
        # rect = self.drawing.bounds
        # return QRectF(rect.origin.x - 1, rect.origin.y - 1, rect.size.width + 2, rect.size.height + 2)


# class DrawingGroupItem(QGraphicsItemGroup):
#     def __init__(self, drawing: DrawingGroup, size: QSizeF, parent: QGraphicsItem = None):
#         super().__init__(parent)
#         self._drawing = drawing
#         self._size = size
#         self._is_active = False

#     @property
#     def drawing(self) -> DrawingGroup:
#         return self._drawing

#     def activate(self) -> None:
#         self._is_active = True

#     def deactivate(self) -> None:
#         self._is_active = False

#     def fresh(self) -> None:
#         for item in self.childItems:
#             self.removeFromGroup(item)
#         for di in self.drawing.children:
#             item = None
#             if isinstance(di, DrawingGroup):
#                 item = DrawingGroupItem(di, self._size)
#             else:
#                 item = DrawingItem(di, self._size)
#             self.addToGroup(item)

#     def paint(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget: Optional[QWidget] = None) -> None:
#         super().paint(painter, option, widget)
#         if self._is_active:
#             painter.setPen(QColor(255, 0, 0))
#             rect = self.drawing.bounds
#             painter.drawRect(converters.qrect(rect))



from __future__ import annotations

from typing import Dict, Optional

import numpy as np

from ImagingS import Point

from . import Transform


class MatrixTransform(Transform):
    S_Matrix = "_matrix"

    def __init__(self, matrix: Optional[np.ndarray] = None) -> None:
        super().__init__()
        self.matrix = matrix if matrix is not None else np.eye(3)

    @property
    def matrix(self) -> np.ndarray:
        return self._matrix

    @matrix.setter
    def matrix(self, value: np.ndarray) -> None:
        assert isinstance(value, np.ndarray)
        assert value.shape == (3, 3)
        self._matrix = value

    def transform(self, origin: Point) -> Point:
        result = Point.fromHomogeneous(
            np.dot(self.matrix, origin.toHomogeneous()))
        return result

    def serialize(self) -> Dict:
        result = super().serialize()

        if "matrix" in result:
            del result["matrix"]

        result[self.S_Matrix] = [*self.matrix[0],
                                 *self.matrix[1], *self.matrix[2]]
        return result

    def deserialize(self, data: Dict) -> None:
        if self.S_Matrix in data:
            mat = data.pop(self.S_Matrix)
            self.matrix = np.array([mat[0:3], mat[3:6], mat[6:9]])
        super().deserialize(data)



import os
import sys

from ImagingS.document import Document

from .instruction import BuiltinInstruction

if __name__ == "__main__":
    input_file = os.path.realpath(sys.argv[1])
    output_dir = os.path.realpath(sys.argv[2])
    os.makedirs(output_dir, exist_ok=True)

    host = BuiltinInstruction(Document(), output_dir)

    with open(input_file, 'r') as fp:
        line = fp.readline()
        while line:
            line = line.strip()
            print(line)
            host.execute(line)
            line = fp.readline()



from abc import ABC, abstractmethod

from ImagingS import Color, IdObject, Point, Rect
from ImagingS.serialization import PropertySerializable


class Brush(PropertySerializable, IdObject, ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def colorAt(self, position: Point, rect: Rect) -> Color: pass



from ImagingS import Color, IdObject, IdObjectList, Point, Rect, Size


def test_color() -> None:
    c = Color(0, 0, 0)
    assert c == Color(0, 0, 0)
    assert c.toHex() == "#000000"
    assert str(c).startswith("Color")


def test_point() -> None:
    c = Point(1, 1)
    assert +c == c.clone()
    assert -c == Point(-1, -1)
    assert c + c == 2 * c
    assert c.asTuple() == (1, 1)
    assert str(c).startswith("Point")
    assert c.toHomogeneous().shape == (3, 1)
    assert c.fromHomogeneous(c.toHomogeneous()) == c


def test_size() -> None:
    c = Size(1, 1)
    assert c == Size(1, 1)
    assert c.asTuple() == (1, 1)
    assert str(c).startswith("Size")


def test_rect() -> None:
    c = Rect(Point(1, 1), Size(1, 1))
    assert c == Rect(Point(1, 1), Size(1, 1))
    assert c == Rect.fromPoints(Point(1, 2), Point(2, 1))
    assert str(c).startswith("Rect")


def test_idobjectlist() -> None:
    ol = IdObjectList()
    obj = IdObject()
    obj.id = "a"
    ol.items = []
    ol.items = ol.items
    assert obj not in ol
    ol.append(obj)
    assert obj in ol
    assert obj.id in ol

    assert len(ol) == 1

    assert ol[0] is obj
    assert ol[obj.id] is obj

    del ol[obj]
    ol.append(obj)
    del ol[0]
    ol.append(obj)
    del ol[obj.id]



from __future__ import annotations

from math import fabs


def feq(a: float, b: float) -> bool:
    return fabs(a-b) < 1e-8


def fsign(a: float) -> int:
    if feq(a, 0):
        return 0
    return 1 if a > 0 else -1



from typing import Any


class Cache:
    def __init__(self) -> None:
        super().__init__()
        self._dict = {}

    def update(self, key: Any, value: Any) -> None:
        self._dict[key] = value

    def outdate(self, key: Any = None) -> None:
        if key is None:
            self._dict.clear()
        elif key in self._dict:
            del self._dict[key]

    def __getitem__(self, key: Any) -> Any:
        if key in self._dict:
            return self._dict[key]
        return None

    def __contains__(self, key: Any) -> bool:
        return key in self._dict



from __future__ import annotations

import json
import lzma
import uuid
from enum import Enum, unique

from ImagingS import IdObject, IdObjectList, Size
from ImagingS.brush import Brush
from ImagingS.drawing import DrawingGroup
from ImagingS.serialization import PropertySerializable
from ImagingS.serialization.json import Decoder, Encoder


@unique
class DocumentFormat(Enum):
    ISD = 1
    RAW = 2


class Document(PropertySerializable, IdObject):
    def __init__(self) -> None:
        super().__init__()
        self.id = str(uuid.uuid1())
        self.brushes = IdObjectList()
        self.drawings = DrawingGroup()
        self.size = Size(600, 600)

    @property
    def brushes(self) -> IdObjectList[Brush]:
        return self._brushes

    @brushes.setter
    def brushes(self, value: IdObjectList[Brush]) -> None:
        assert isinstance(value, IdObjectList)
        self._brushes = value

    @property
    def drawings(self) -> DrawingGroup:
        return self._drawings

    @drawings.setter
    def drawings(self, value: DrawingGroup) -> None:
        assert isinstance(value, DrawingGroup)
        self._drawings = value

    @property
    def size(self) -> Size:
        return self._size

    @size.setter
    def size(self, value: Size) -> None:
        assert isinstance(value, Size)
        self._size = value

    def save(self, file, format: DocumentFormat = DocumentFormat.ISD) -> None:
        if format is DocumentFormat.RAW:
            json.dump(self, file, ensure_ascii=False, indent=4, cls=Encoder)
        else:
            s = json.dumps(self, ensure_ascii=False, cls=Encoder)
            data = s.encode("utf8")
            data = lzma.compress(data)
            file.write(data)

    @staticmethod
    def load(file, format: DocumentFormat = DocumentFormat.ISD) -> Document:
        if format is DocumentFormat.RAW:
            result = json.load(file, cls=Decoder)
        else:
            data = file.read()
            data = lzma.decompress(data)
            s = data.decode("utf8")
            result = json.loads(s, cls=Decoder)
        assert isinstance(result, Document)
        return result



from ._Transform import Transform, TransformGroup
from ._Matrix import MatrixTransform
from ._Rotate import RotateTransform
from ._Scale import ScaleTransform
from ._Skew import SkewTransform
from ._Translate import TranslateTransform

__all__ = (
    "Transform",
    "TransformGroup",
    "MatrixTransform",
    "RotateTransform",
    "ScaleTransform",
    "SkewTransform",
    "TranslateTransform",
)



from math import floor
from typing import Union

from PyQt5.QtCore import QPointF, Qt
from PyQt5.QtGui import QKeyEvent, QPainterPath

from ImagingS.drawing import DrawingGroup
from ImagingS.geometry import Geometry
from ImagingS.Gui import converters
from ImagingS.transform import ScaleTransform

from . import TransformInteractivity


class ScaleTransformInteractivity(TransformInteractivity[ScaleTransform]):
    def __init__(self, target: Union[DrawingGroup, Geometry], transform: ScaleTransform) -> None:
        super().__init__(target, transform)

    def start(self) -> None:
        self._hasStarted = False
        self._isShift = False
        super().start()

    def _updateData(self) -> None:
        D = 40  # point distance
        BOUND = 3.0

        delta = self._end - self._start
        fx, fy = (max(min(delta.x / D, BOUND), -BOUND),
                  max(min(delta.y / D, BOUND), -BOUND))
        rfx, rfy = fx, fy
        if rfx < 0:
            rfx = 1.0 / (1+abs(rfx))
        else:
            rfx += 1
        if rfy < 0:
            rfy = 1.0 / (1+abs(rfy))
        else:
            rfy += 1
        self.transform.factor = rfx, rfy

        path = QPainterPath()
        s = converters.qpoint(self._start)
        t = QPointF(s.x()+fx*D, s.y()+fy*D)
        fbound = floor(BOUND)
        for i in range(1, fbound + 1):
            path.addEllipse(QPointF(s.x()+i*D, s.y()),
                            (fbound + i)/2, (fbound + i)/2)
            path.addEllipse(QPointF(s.x()-i*D, s.y()),
                            (fbound - i)/2, (fbound - i)/2)
            path.addEllipse(QPointF(s.x(), s.y()+i*D),
                            (fbound + i)/2, (fbound + i)/2)
            path.addEllipse(QPointF(s.x(), s.y()-i*D),
                            (fbound - i)/2, (fbound - i)/2)

        tx = QPointF(s.x()+fx*D, s.y())
        ty = QPointF(s.x(), s.y()+fy*D)
        path.addEllipse(s, 3, 3)
        path.addEllipse(t, 3, 3)

        path.moveTo(s)
        path.lineTo(QPointF(s.x()+BOUND*D, s.y()))
        path.moveTo(s)
        path.lineTo(QPointF(s.x()-BOUND*D, s.y()))
        path.moveTo(s)
        path.lineTo(QPointF(s.x(), s.y()+BOUND*D))
        path.moveTo(s)
        path.lineTo(QPointF(s.x(), s.y()-BOUND*D))
        path.moveTo(t)
        path.lineTo(tx)
        path.moveTo(t)
        path.lineTo(ty)
        self._viewItem.setPath(path)

    def _setEndPoint(self, point: QPointF) -> None:
        if self._isShift:
            tp = converters.point(point)
            delta = tp - self._start
            if abs(delta.x) <= abs(delta.y):  # equal factor(x,y) are same
                delta.y = delta.x
            else:
                delta.x = delta.y
            self._end = self._start + delta
        else:
            self._end = converters.point(point)

    def onMouseRelease(self, point: QPointF) -> None:
        if not self._hasStarted:
            self._start = converters.point(point)
            self.transform.center = self._start
            self._hasStarted = True
        else:
            self._setEndPoint(point)
            self._updateData()
            self.end(True)
        super().onMouseRelease(point)

    def onMouseMove(self, point: QPointF) -> None:
        if self._hasStarted:
            self._setEndPoint(point)
            self._updateData()
            self.update()
        super().onMouseMove(point)

    def onKeyPress(self, key: QKeyEvent) -> None:
        if key.key() == Qt.Key_Shift:
            self._isShift = True
        super().onKeyPress(key)

    def onKeyRelease(self, key: QKeyEvent) -> None:
        if key.key() == Qt.Key_Shift:
            self._isShift = False
        super().onKeyRelease(key)



from __future__ import annotations

from enum import Enum, unique
from typing import Optional

from PyQt5.QtCore import QObject, QPointF, pyqtSignal
from PyQt5.QtGui import QKeyEvent
from PyQt5.QtWidgets import QGraphicsItem


@unique
class InteractivityState(Enum):
    Pending = 1,
    Running = 2,
    Success = 3,
    Failed = 4


class Interactivity(QObject):
    started = pyqtSignal(QObject)
    ended = pyqtSignal(QObject)
    updated = pyqtSignal(QObject)

    def __init__(self) -> None:
        super().__init__()
        self._isNeedRender = False
        self._state = InteractivityState.Pending

    def start(self) -> None:  # will emit event, so call at the end in subclass
        self._state = InteractivityState.Running
        self.started.emit(self)

    def end(self, success: bool) -> None:
        if success:
            self._state = InteractivityState.Success
        else:
            self._state = InteractivityState.Failed
        self.ended.emit(self)

    def update(self) -> None:
        self.updated.emit(self)

    @property
    def viewItem(self) -> Optional[QGraphicsItem]:
        return None

    @property
    def state(self) -> InteractivityState:
        return self._state

    def onMousePress(self, point: QPointF) -> None: pass

    def onMouseMove(self, point: QPointF) -> None: pass

    def onMouseRelease(self, point: QPointF) -> None: pass

    def onMouseDoubleClick(self, point: QPointF) -> None: pass

    def onKeyPress(self, key: QKeyEvent) -> None: pass

    def onKeyRelease(self, key: QKeyEvent) -> None: pass



import qtawesome as qta
from PyQt5.QtGui import QIcon

from ImagingS import Color
from ImagingS.brush import SolidBrush

from . import converters

grid: QIcon = qta.icon("mdi.grid")

refresh: QIcon = qta.icon("mdi.refresh")

list: QIcon = qta.icon("mdi.format-list-numbered")

set: QIcon = qta.icon("mdi.code-braces")

code: QIcon = qta.icon("mdi.code-tags")

fileTree: QIcon = qta.icon("mdi.file-tree")

visual: QIcon = qta.icon("mdi.image")

vertex: QIcon = qta.icon("mdi.vector-point")

stroke: QIcon = qta.icon("mdi.border-color")

fill: QIcon = qta.icon("mdi.format-color-fill")

x: QIcon = qta.icon("mdi.axis-x-arrow")

y: QIcon = qta.icon("mdi.axis-y-arrow")

r: QIcon = qta.icon("mdi.alpha-r-circle", color="red")

g: QIcon = qta.icon("mdi.alpha-g-circle", color="green")

b: QIcon = qta.icon("mdi.alpha-b-circle", color="blue")

width: QIcon = qta.icon("mdi.alpha-w-circle")

height: QIcon = qta.icon("mdi.alpha-h-circle")

id: QIcon = qta.icon("mdi.identifier")

angle: QIcon = qta.icon("mdi.angle-acute")

matrix: QIcon = qta.icon("mdi.matrix")

center: QIcon = qta.icon("mdi.image-filter-center-focus")

document: QIcon = qta.icon("mdi.file-document")

size: QIcon = qta.icon("mdi.format-size")

rect: QIcon = qta.icon("mdi.rectangle-outline")

point: QIcon = qta.icon("mdi.circle-medium")

drawing: QIcon = qta.icon("mdi.drawing-box")

geometry: QIcon = qta.icon("mdi.drawing")

brush: QIcon = qta.icon("mdi.brush")

solidBrush: QIcon = qta.icon("mdi.solid")

property: QIcon = qta.icon("mdi.database")

transform: QIcon = qta.icon("mdi.flash")

lineGeometry: QIcon = qta.icon("mdi.vector-line")

curveGeometry: QIcon = qta.icon("mdi.vector-curve")

ellipseGeometry: QIcon = qta.icon("mdi.vector-ellipse")

polygonGeometry: QIcon = qta.icon("mdi.vector-polygon")

polylineGeometry: QIcon = qta.icon("mdi.vector-polyline")

rectangleGeometry: QIcon = qta.icon("mdi.vector-rectangle")

clip: QIcon = qta.icon("mdi.crop")

skewTransform: QIcon = qta.icon("mdi.skew-more")

scaleTransform: QIcon = qta.icon("mdi.relative-scale")

translateTransform: QIcon = qta.icon("mdi.cursor-move")

rotateTransform: QIcon = qta.icon("mdi.rotate-left")

matrixTransform: QIcon = matrix

transformGroup: QIcon = qta.icon("mdi.group")


def getColorIcon(color: Color) -> QIcon:
    return qta.icon("mdi.invert-colors", color=converters.qcolor(color))


def getBrushIcon(brush: SolidBrush) -> QIcon:
    # pixmap = QPixmap(32, 32)
    # pixmap.fill(converters.convert_color(color))
    # return QIcon(pixmap)
    return qta.icon("mdi.brush", color=converters.qcolor(brush.color))



from __future__ import annotations

from enum import Enum, unique
from typing import Iterable, Iterator, Optional, cast

from ImagingS import Point, Rect, fsign
from ImagingS.drawing import Pen

from . import Geometry


@unique
class LineAlgorithm(Enum):
    Dda = 0
    Bresenham = 1


@unique
class LineClipAlgorithm(Enum):
    CohenSutherland = 0
    LiangBarsky = 1


def _genDDA(start: Point, end: Point) -> Iterator[Point]:
    if start == end:
        yield start
        return
    xs, ys = map(round, start.asTuple())
    xe, ye = map(round, end.asTuple())
    dx, dy = xe-xs, ye-ys
    sx, sy = fsign(dx), fsign(dy)
    dx, dy = abs(dx), abs(dy)
    flag = False
    if dy > dx:
        dx, dy = dy, dx
        flag = True
    k = dy / dx
    if flag:
        k *= sx
    else:
        k *= sy
    xc, yc = xs, ys
    for _ in range(round(abs(dx))+1):
        yield Point(xc, yc)
        if flag:
            yc += sy
            xc += k
        else:
            xc += sx
            yc += k


def _genBresenham(start: Point, end: Point) -> Iterator[Point]:
    if start == end:
        yield start
        return
    xs, ys = map(round, start.asTuple())
    xe, ye = map(round, end.asTuple())
    dx, dy = xe-xs, ye-ys
    sx, sy = fsign(dx), fsign(dy)
    dx, dy = abs(dx), abs(dy)
    flag = False
    if dy > dx:
        dx, dy = dy, dx
        flag = True
    xc, yc = xs, ys
    ne = 2*dy-dx
    for _ in range(round(dx)+1):
        yield Point(xc, yc)
        if ne >= 0:
            if flag:
                xc += sx
            else:
                yc += sy
            ne -= 2*dx
        if flag:
            yc += sy
        else:
            xc += sx
        ne += 2*dy


class LineGeometry(Geometry):
    def __init__(self, start: Optional[Point] = None, end: Optional[Point] = None, algorithm: Optional[LineAlgorithm] = None) -> None:
        super().__init__()
        self.start = start if start else Point()
        self.end = end if end else Point()
        self.algorithm = algorithm if algorithm else LineAlgorithm.Dda

    @property
    def start(self) -> Point:
        return self._start

    @start.setter
    def start(self, value: Point) -> None:
        assert isinstance(value, Point)
        self._start = value
        self.refreshBounds()

    @property
    def end(self) -> Point:
        return self._end

    @end.setter
    def end(self, value: Point) -> None:
        assert isinstance(value, Point)
        self._end = value
        self.refreshBounds()

    @property
    def algorithm(self) -> LineAlgorithm:
        return self._algorithm

    @algorithm.setter
    def algorithm(self, value: LineAlgorithm) -> None:
        assert isinstance(value, LineAlgorithm)
        self._algorithm = value

    def transformed(self) -> Geometry:
        if self.transform is None:
            return self
        else:
            start = self.transform.transform(self.start)
            end = self.transform.transform(self.end)
            return LineGeometry(start, end, self.algorithm)

    def strokePoints(self, pen: Pen) -> Iterable[Point]:
        target = self
        if self.transform is not None:
            target = cast(LineGeometry, self.transformed())

        gen = None
        if target.algorithm is LineAlgorithm.Dda:
            gen = _genDDA
        elif target.algorithm is LineAlgorithm.Bresenham:
            gen = _genBresenham
        return gen(target.start, target.end)

    def fillPoints(self) -> Iterable[Point]:
        return []

    def _calculateBounds(self) -> Rect:
        target = self
        if self.transform is not None:
            target = cast(LineGeometry, self.transformed())
        return Rect.fromPoints(target.start, target.end)



from PyQt5.QtGui import QPainter

from ImagingS import Color, Point, Rect
from ImagingS.drawing import RenderContext
from ImagingS.Gui import converters


class PainterRenderContext(RenderContext):
    def __init__(self, painter: QPainter, bounds: Rect):
        super().__init__()
        self._painter = painter
        self._bounds = bounds

    def _point(self, position: Point, color: Color) -> None:
        self._painter.setPen(converters.qcolor(color))
        self._painter.drawPoint(converters.qpoint(position))

    def bounds(self) -> Rect:
        return self._bounds



import math
from typing import Union

from PyQt5.QtCore import QPointF, QRectF, Qt
from PyQt5.QtGui import QKeyEvent, QPainterPath

from ImagingS import Point, fsign
from ImagingS.drawing import DrawingGroup
from ImagingS.geometry import Geometry
from ImagingS.Gui import converters
from ImagingS.transform import SkewTransform

from . import TransformInteractivity


class SkewTransformInteractivity(TransformInteractivity[SkewTransform]):
    def __init__(self, target: Union[DrawingGroup, Geometry], transform: SkewTransform) -> None:
        super().__init__(target, transform)

    def start(self) -> None:
        self._hasStarted = False
        self._isShift = False
        super().start()

    def _updateData(self) -> None:  # y-axis: zero rad
        R = 50
        BOUND = math.pi / 2 - math.pi / 6
        ax, ay = self.transform.angle

        if hasattr(self, "_second"):
            delta = self._second - self._center
            ay = abs(math.atan2(abs(delta.x), abs(delta.y))) * \
                fsign(delta.x)  # x-axis positive means > 0
            ay = max(min(ay, BOUND), -BOUND)
        else:
            delta = self._first - self._center
            ax = abs(math.atan2(abs(delta.x), abs(delta.y))) * \
                fsign(delta.x)  # x-axis positive means > 0
            ax = max(min(ax, BOUND), -BOUND)

        self.transform.angle = (ax, ay)

        path = QPainterPath()
        s = converters.qpoint(self._center)
        rect = QRectF(s.x() - R, s.y() - R, 2*R, 2*R)
        dax, day = ax / math.pi * 180, ay / math.pi * 180
        path.moveTo(s)
        path.arcTo(rect, 90.0, -dax)  # x-axis positive means > 0
        path.lineTo(s)
        path.arcTo(rect, -90.0, day)  # x-axis positive means > 0
        path.lineTo(s)
        self._viewItem.setPath(path)

    def _setPoint(self, point: QPointF) -> Point:
        if self._isShift:
            tp = converters.point(point)
            delta = tp - self._center
            if abs(delta.x) <= abs(delta.y):
                delta.x = 0
            else:
                delta.y = 0
            return self._center + delta
        else:
            return converters.point(point)

    def onMouseRelease(self, point: QPointF) -> None:
        if not self._hasStarted:
            self._center = converters.point(point)
            self.transform.center = self._center
            self._first = self._center
            self._hasStarted = True
        else:
            if hasattr(self, "_second"):
                self._second = self._setPoint(point)
                self._updateData()
                self.end(True)
            else:
                self._first = self._setPoint(point)
                self._second = self._center
                self._updateData()
        super().onMouseRelease(point)

    def onMouseMove(self, point: QPointF) -> None:
        if self._hasStarted:
            if hasattr(self, "_second"):
                self._second = self._setPoint(point)
            else:
                self._first = self._setPoint(point)
            self._updateData()
            self.update()
        super().onMouseMove(point)

    def onKeyPress(self, key: QKeyEvent) -> None:
        if key.key() == Qt.Key_Shift:
            self._isShift = True
        super().onKeyPress(key)

    def onKeyRelease(self, key: QKeyEvent) -> None:
        if key.key() == Qt.Key_Shift:
            self._isShift = False
        super().onKeyRelease(key)



from ImagingS import Point, Rect, Size
from ImagingS.drawing import Pen
from ImagingS.geometry import (CurveAlgorithm, CurveGeometry, EllipseGeometry,
                               GeometryGroup, LineAlgorithm, LineGeometry,
                               PolygonGeometry, PolylineGeometry,
                               RectangleGeometry)


def test_line() -> None:
    line = LineGeometry(Point(
        0, 0), Point(1, 1), LineAlgorithm.Dda)
    assert line.inStroke(Pen(), Point())
    assert not line.inFill(Point())
    line = LineGeometry(Point(
        0, 0), Point(1, 1), LineAlgorithm.Bresenham)
    assert line.inStroke(Pen(), Point())
    assert not line.inFill(Point())


def test_poly() -> None:
    vertex = [Point(0, 0), Point(
        1, 0), Point(1, 1), Point(0.5, 0.5)]
    line = PolylineGeometry(vertex, LineAlgorithm.Dda)
    assert line.inStroke(Pen(), Point())
    assert not line.inFill(Point())
    line = PolygonGeometry(vertex, LineAlgorithm.Dda)
    assert line.inStroke(Pen(), Point())
    assert not line.inFill(Point())
    line = RectangleGeometry(
        Rect(Point(), Size(10, 10)), LineAlgorithm.Dda)
    assert line.inStroke(Pen(), Point())
    assert not line.inFill(Point())


def test_curve() -> None:
    control = [Point(0, 0), Point(
        1, 0), Point(1, 1), Point(0.5, 0.5)]
    curve = CurveGeometry(control, CurveAlgorithm.Bezier)
    assert curve.inStroke(Pen(), Point())
    curve = CurveGeometry(control, CurveAlgorithm.BSpline)
    assert not curve.inStroke(Pen(), Point())


def test_ellipse() -> None:
    ell = EllipseGeometry.fromRect(
        Rect(Point(), Size(10, 10)))
    assert not ell.inStroke(Pen(), Point())


def test_group() -> None:
    line = LineGeometry(Point(
        0, 0), Point(1, 1), LineAlgorithm.Dda)
    ell = EllipseGeometry.fromRect(
        Rect(Point(), Size(10, 10)))
    group = GeometryGroup()
    group.children.append(line)
    group.children.append(ell)
    list(group.strokePoints(Pen()))
    list(group.fillPoints())
    assert group.inStroke(Pen(), Point())
    assert not group.inFill(Point())



from __future__ import annotations

from math import cos, sin
from typing import Dict, Optional

import numpy as np

from ImagingS import Point
from ImagingS.transform import MatrixTransform


class RotateTransform(MatrixTransform):
    def __init__(self, center: Optional[Point] = None, angle: float = 0.0) -> None:
        super().__init__()
        self.center = center if center else Point()
        self.angle = angle

    @property
    def center(self) -> Point:
        return self._center

    @center.setter
    def center(self, value: Point) -> None:
        assert isinstance(value, Point)
        self._center = value

    @property
    def angle(self) -> float:
        return self._angle

    @angle.setter
    def angle(self, value: float) -> None:
        self._angle = float(value)
        self.matrix = np.array(
            [[cos(self._angle), -sin(self._angle), 0],
             [sin(self._angle), cos(self._angle), 0],
             [0, 0, 1]])

    def transform(self, origin: Point) -> Point:
        return super().transform(origin - self.center) + self.center

    def serialize(self) -> Dict:
        result = super().serialize()
        if MatrixTransform.S_Matrix in result:
            del result[MatrixTransform.S_Matrix]
        return result



import sys


def main():
    from .app import Application
    app = Application(sys.argv)

    from .widgets import MainWindow

    mainWindow = MainWindow()
    mainWindow.show()
    sys.exit(app.run())


if __name__ == "__main__":
    main()



from enum import Enum
from typing import Any, Dict, List

import qtawesome as qta
from PyQt5.QtCore import QModelIndex, Qt, pyqtSignal
from PyQt5.QtGui import QIcon, QStandardItem, QStandardItemModel

from ImagingS import Color
from ImagingS.Gui import icons
from ImagingS.serialization import Property, getProperties


class PropertyModel(QStandardItemModel):
    NAME, VALUE = range(2)
    changed = pyqtSignal()

    def __init__(self, parent) -> None:
        super().__init__(0, 2, parent)
        self.setHeaderData(PropertyModel.NAME, Qt.Horizontal, "Name")
        self.setHeaderData(PropertyModel.VALUE, Qt.Horizontal, "Value")
        self.onlyEditable = False
        self.obj = None

    def fresh(self, obj: Any = None) -> None:
        self.removeRows(0, self.rowCount())
        self.obj = obj
        if obj is None:
            return
        name = obj.__class__.__name__
        item = QStandardItem(name)
        item.setEditable(False)
        self._setIcon(item, name, obj)
        self._setUserData(item, obj)
        self.appendRow(item)
        self._addPropChildren(item, obj)

    @property
    def onlyEditable(self) -> bool:
        return self._onlyEditable

    @onlyEditable.setter
    def onlyEditable(self, value: bool) -> None:
        self._onlyEditable = value

    def data(self, index: QModelIndex, role: int) -> Any:
        if index.column() == self.VALUE:
            value = self.getUserData(index)
            if isinstance(value, Property):
                realValue = value.get()
                if role == Qt.DisplayRole or role == Qt.EditRole:
                    if realValue is None:
                        return "None"
                    elif isinstance(realValue, str):
                        return realValue
                    elif isinstance(realValue, int):
                        return str(realValue)
                    elif isinstance(realValue, float):
                        return str(round(realValue, 2))
                    elif isinstance(realValue, Enum):
                        return realValue.name
        return super().data(index, role)

    def setData(self, index: QModelIndex, value: Any, role: int) -> bool:
        if index.column() == self.VALUE:
            data = self.getUserData(index)
            if isinstance(data, Property):
                realData = data.get()
                if role == Qt.DisplayRole or role == Qt.EditRole:
                    modified = True
                    if isinstance(realData, str):
                        try:
                            data.set(str(value))
                        except Exception:
                            return False
                    elif isinstance(realData, int):
                        try:
                            data.set(int(value))
                        except Exception:
                            return False
                    elif isinstance(realData, float):
                        try:
                            data.set(float(value))
                        except Exception:
                            return False
                    elif isinstance(realData, Enum):
                        try:
                            en = getattr(realData.__class__, value)
                            assert isinstance(en, realData.__class__)
                            data.set(en)
                        except Exception:
                            return False
                    else:
                        modified = False
                    if modified:
                        self.dataChanged.emit(index, index, [role])
                        self.changed.emit()
                        return True
        return super().setData(index, value, role)

    def getUserData(self, index: QModelIndex) -> Any:
        item = self.itemFromIndex(index)
        return item.data(Qt.UserRole)

    def _setUserData(self, item: QStandardItem, value: Any) -> None:
        item.setData(value, Qt.UserRole)

    def _setIcon(self, item: QStandardItem, name: str, value: Any) -> None:
        if hasattr(icons, name):
            icon = getattr(icons, name)
            if isinstance(icon, QIcon):
                item.setIcon(icon)
                return
        typeName = value.__class__.__name__
        typeName = typeName[0].lower() + typeName[1:]
        if hasattr(icons, typeName):
            icon = getattr(icons, typeName)
            item.setIcon(icon)
            return
        if name == "vertexes" or name == "controlPoints":
            item.setIcon(icons.vertex)
        elif name == "brushes":
            item.setIcon(icons.brush)
        elif name == "drawings":
            item.setIcon(icons.drawing)
        elif name == "algorithm":
            item.setIcon(qta.icon("mdi.lightbulb"))
        elif isinstance(value, Color):
            item.setIcon(icons.getColorIcon(value))

    def _addDictChildren(self, root: QStandardItem, li: Dict) -> None:
        for k, v in li.items():
            self._addChild(root, str(k), v)

    def _addListChildren(self, root: QStandardItem, li: List) -> None:
        for i, value in enumerate(li):
            self._addChild(root, str(i), value)

    def _addPropChildren(self, root: QStandardItem, obj: Any) -> bool:  # return has child
        flag = False
        for prop in getProperties(obj):
            if self.onlyEditable and not prop.canSet:
                continue
            flag = True
            self._addChild(root, prop.name, prop)
        return flag

    def _addChild(self, root: QStandardItem, name: str, value: Any) -> None:
        item = QStandardItem(name)
        item.setEditable(False)
        self._setUserData(item, value)
        valueItem = QStandardItem()
        self._setUserData(valueItem, value)

        realValue = value
        if isinstance(value, Property):
            realValue = value.get()
            valueItem.setEditable(value.canSet)

        self._setIcon(item, name, realValue)
        root.appendRow(item)
        index = self.indexFromItem(item).row()
        if realValue is None or isinstance(realValue, int) or isinstance(realValue, float) or isinstance(realValue, str):
            pass
        elif isinstance(realValue, list):
            valueItem.setText("List")
            valueItem.setEditable(False)
            self._addListChildren(item, realValue)
        elif isinstance(realValue, dict):
            valueItem.setText("Dictionary")
            valueItem.setEditable(False)
            self._addDictChildren(item, realValue)
        elif isinstance(realValue, set):
            valueItem.setText("Set")
            valueItem.setEditable(False)
            self._addListChildren(item, list(realValue))
        else:
            if self._addPropChildren(item, realValue):
                valueItem.setText(realValue.__class__.__name__)
                valueItem.setEditable(False)
            else:
                valueItem.setText(str(realValue))
                valueItem.setEditable(False)
        if name == "id":
            valueItem.setEditable(False)
        root.setChild(index, self.VALUE, valueItem)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/runner/work/ImagingS/ImagingS/src/ImagingS/Gui/uic/VisualPage.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_VisualPage(object):
    def setupUi(self, VisualPage):
        VisualPage.setObjectName("VisualPage")
        VisualPage.resize(530, 300)
        self.horizontalLayout = QtWidgets.QHBoxLayout(VisualPage)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.tlbMain = QtWidgets.QToolBar(VisualPage)
        self.tlbMain.setOrientation(QtCore.Qt.Vertical)
        self.tlbMain.setObjectName("tlbMain")
        self.horizontalLayout.addWidget(self.tlbMain)
        self.widMain = QtWidgets.QWidget(VisualPage)
        self.widMain.setObjectName("widMain")
        self.grdMain = QtWidgets.QGridLayout(self.widMain)
        self.grdMain.setContentsMargins(0, 0, 0, 0)
        self.grdMain.setObjectName("grdMain")
        self.horizontalLayout.addWidget(self.widMain)
        self.actDrawingLine = QtWidgets.QAction(VisualPage)
        self.actDrawingLine.setObjectName("actDrawingLine")
        self.actDrawingCurve = QtWidgets.QAction(VisualPage)
        self.actDrawingCurve.setObjectName("actDrawingCurve")
        self.actDrawingEllipse = QtWidgets.QAction(VisualPage)
        self.actDrawingEllipse.setObjectName("actDrawingEllipse")
        self.actTransformTranslate = QtWidgets.QAction(VisualPage)
        self.actTransformTranslate.setObjectName("actTransformTranslate")
        self.actTransformScale = QtWidgets.QAction(VisualPage)
        self.actTransformScale.setObjectName("actTransformScale")
        self.actTransformRotate = QtWidgets.QAction(VisualPage)
        self.actTransformRotate.setObjectName("actTransformRotate")
        self.actTransformSkew = QtWidgets.QAction(VisualPage)
        self.actTransformSkew.setObjectName("actTransformSkew")
        self.actTransformMatrix = QtWidgets.QAction(VisualPage)
        self.actTransformMatrix.setObjectName("actTransformMatrix")
        self.actTransformGroup = QtWidgets.QAction(VisualPage)
        self.actTransformGroup.setObjectName("actTransformGroup")
        self.actDrawingPolygon = QtWidgets.QAction(VisualPage)
        self.actDrawingPolygon.setObjectName("actDrawingPolygon")
        self.actDrawingPolyline = QtWidgets.QAction(VisualPage)
        self.actDrawingPolyline.setObjectName("actDrawingPolyline")
        self.actDrawingRectangle = QtWidgets.QAction(VisualPage)
        self.actDrawingRectangle.setObjectName("actDrawingRectangle")
        self.actClip = QtWidgets.QAction(VisualPage)
        self.actClip.setObjectName("actClip")
        self.tlbMain.addAction(self.actDrawingLine)
        self.tlbMain.addAction(self.actDrawingPolyline)
        self.tlbMain.addAction(self.actDrawingPolygon)
        self.tlbMain.addAction(self.actDrawingRectangle)
        self.tlbMain.addAction(self.actDrawingCurve)
        self.tlbMain.addAction(self.actDrawingEllipse)
        self.tlbMain.addSeparator()
        self.tlbMain.addAction(self.actTransformTranslate)
        self.tlbMain.addAction(self.actTransformScale)
        self.tlbMain.addAction(self.actTransformRotate)
        self.tlbMain.addAction(self.actTransformSkew)
        self.tlbMain.addAction(self.actTransformMatrix)
        self.tlbMain.addAction(self.actTransformGroup)
        self.tlbMain.addSeparator()
        self.tlbMain.addAction(self.actClip)

        self.retranslateUi(VisualPage)
        QtCore.QMetaObject.connectSlotsByName(VisualPage)

    def retranslateUi(self, VisualPage):
        _translate = QtCore.QCoreApplication.translate
        VisualPage.setWindowTitle(_translate("VisualPage", "Visual"))
        self.tlbMain.setWindowTitle(_translate("VisualPage", "Visual Tool Box"))
        self.actDrawingLine.setText(_translate("VisualPage", "Line"))
        self.actDrawingCurve.setText(_translate("VisualPage", "Curve"))
        self.actDrawingEllipse.setText(_translate("VisualPage", "Ellipse"))
        self.actTransformTranslate.setText(_translate("VisualPage", "Translate"))
        self.actTransformScale.setText(_translate("VisualPage", "Scale"))
        self.actTransformRotate.setText(_translate("VisualPage", "Rotate"))
        self.actTransformSkew.setText(_translate("VisualPage", "Skew"))
        self.actTransformMatrix.setText(_translate("VisualPage", "Matrix"))
        self.actTransformGroup.setText(_translate("VisualPage", "Group"))
        self.actDrawingPolygon.setText(_translate("VisualPage", "Polygon"))
        self.actDrawingPolyline.setText(_translate("VisualPage", "Polyline"))
        self.actDrawingRectangle.setText(_translate("VisualPage", "Rectangle"))
        self.actClip.setText(_translate("VisualPage", "Clip"))



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional

from ImagingS import Color, IdObject, IdObjectList, Point, Rect, RectMeasurer
from ImagingS.serialization import PropertySerializable
from ImagingS.transform import Transform

from . import ClipRenderContext, ProxyRenderContext, RenderContext


class Drawing(PropertySerializable, IdObject, ABC):
    def __init__(self) -> None:
        super().__init__()
        self.setParent(None)

    @abstractmethod
    def render(self, context: RenderContext) -> None: pass

    @property
    @abstractmethod
    def bounds(self) -> Rect: pass


class DrawingGroup(Drawing):
    def __init__(self) -> None:
        super().__init__()
        self.children = IdObjectList()
        self.transform = None
        self.clip = None

    @property
    def children(self) -> IdObjectList[Drawing]:
        return self._children

    @children.setter
    def children(self, value: IdObjectList[Drawing]) -> None:
        assert isinstance(value, IdObjectList)
        self._children = value
        self.refreshBounds()

    @property
    def transform(self) -> Optional[Transform]:
        return self._transform

    @transform.setter
    def transform(self, value: Optional[Transform]) -> None:
        assert isinstance(value, (type(None), Transform))
        self._transform = value
        self.refreshBounds()

    @property
    def clip(self) -> Optional[Rect]:
        return self._clip

    @clip.setter
    def clip(self, value: Optional[Rect]) -> None:
        assert isinstance(value, (type(None), Rect))
        self._clip = value

    def render(self, context: RenderContext) -> None:
        renderContext = context
        if self.transform is not None:
            def fpoint(position: Point, color: Color) -> None:
                position = self.transform.transform(position)
                context.point(position, color)

            def fbounds() -> Rect:
                return context.bounds()

            renderContext = ProxyRenderContext(fpoint, fbounds)

        if self.clip is not None:
            renderContext = ClipRenderContext(self.clip, renderContext)

        for item in self.children:
            item.render(renderContext)

    @property
    def bounds(self) -> Rect:
        if not hasattr(self, "_bounds"):
            if hasattr(self, "_measurering"):
                return Rect()
            self._measurering = True
            self._bounds = self._calculateBounds()
            del self._measurering
        return self._bounds

    def _calculateBounds(self) -> Rect:
        measurer = RectMeasurer()

        def fpoint(position: Point, color: Color) -> None:
            measurer.append(position)

        def frect() -> Rect:
            return Rect.infinite()

        context = ProxyRenderContext(fpoint, frect)

        self.render(context)
        return measurer.result()

    def refreshBounds(self) -> None:
        if hasattr(self, "_bounds"):
            del self._bounds
